"""
Training
--------

.. autofunction:: nussl.ml.train.create_train_and_validation_engines

.. autofunction:: nussl.ml.train.add_tensorboard_handler

.. autofunction:: nussl.ml.train.cache_dataset

.. autofunction:: nussl.ml.train.add_validate_and_checkpoint

.. autofunction:: nussl.ml.train.add_stdout_handler

.. autofunction:: nussl.ml.train.add_progress_bar_handler

.. autoclass:: nussl.ml.train.ValidationEvents
    :members:
    :undoc-members:

.. autoclass:: nussl.ml.train.BackwardsEvents
    :members:
    :undoc-members:

Loss functions
--------------

.. automodule:: nussl.ml.train.loss
    :members:
    :undoc-members:
    :autosummary:

Closures
--------

.. automodule:: nussl.ml.train.closures
    :members:
    :undoc-members:
    :autosummary:

"""

from .trainer import (
    create_train_and_validation_engines,
    add_tensorboard_handler,
    cache_dataset,
    add_validate_and_checkpoint,
    add_stdout_handler,
    add_progress_bar_handler,
    ValidationEvents,
    BackwardsEvents
)

from . import loss
from . import closures



"""
Core
====

AudioSignals
------------
.. autoclass:: nussl.core.AudioSignal
    :members:
    :autosummary:

Masks
-----
.. automodule:: nussl.core.masks
    :members:
    :autosummary:

Constants
------------
.. automodule:: nussl.core.constants
    :members:
    :autosummary:

External File Zoo
-----------------
.. automodule:: nussl.core.efz_utils
    :members:
    :autosummary:


General utilities
-----------------
.. automodule:: nussl.core.utils
    :members:
    :autosummary:

Audio effects
-------------
.. automodule:: nussl.core.effects
    :members:
    :autosummary:

Mixing
------
.. automodule:: nussl.core.mixing
    :members:
    :autosummary:

Playing and embedding audio
---------------------------
.. automodule:: nussl.core.play_utils
    :members:
    :autosummary:

Checkpoint migration (backwards compatability)
----------------------------------------------
.. automodule:: nussl.core.migration
    :members:
    :autosummary:

"""

from .audio_signal import AudioSignal, STFTParams
from . import constants
from . import efz_utils
from . import play_utils
from . import utils
from . import mixing
from . import masks

__all__ = [
    'AudioSignal', 
    'STFTParams',
    'constants',
    'efz_utils',
    'play_utils',
    'utils',
    'mixing'
    'masks',
]



import torch
from ..base import SeparationBase, DeepMixin, SeparationException


class DeepAudioEstimation(DeepMixin, SeparationBase):
    """
    Separates an audio signal using a model that produces separated sources directly
    in the waveform domain. It expects that the model outputs a dictionary where one
    of the keys is 'audio'. This uses the `DeepMixin` class to load the model
    and set the audio signal's parameters to be appropriate for the model.
    
    Args:
        input_audio_signal: (AudioSignal`) An AudioSignal object containing the 
          mixture to be separated.        
        model_path (str, optional): Path to the model that will be used. Can be None, 
          so that you can initialize a class and load the model later.  
          Defaults to None.
        extra_data: A dictionary containing any additional data that will 
          be merged with the output dictionary.
        device (str, optional): Device to put the model on. Defaults to 'cpu'.
        **kwargs (dict): Keyword arguments for MaskSeparationBase.
    """
    def __init__(self, input_audio_signal, model_path=None, device='cpu', 
                 extra_data=None, **kwargs):
        super().__init__(input_audio_signal, **kwargs)
        if model_path is not None:
            self.load_model(model_path, device=device)
        self.model_output = None
        self.extra_data = extra_data
        # audio channel dimension in an audio model
        self.channel_dim = 1

    def forward(self):
        input_data = self._get_input_data_for_model(self.extra_data)
        with torch.no_grad():
            output = self.model(input_data)
            if 'audio' not in output:
                raise SeparationException(
                    "This model is not a deep audio estimation model! "
                    "Did not find 'audio' key in output dictionary.")
            audio = output['audio']
            # swap back batch and sample dims
            if self.metadata['num_channels'] == 1:
                audio = audio.transpose(0, self.channel_dim)
            audio = audio.squeeze(0)
            audio = audio.cpu().data.numpy()
        self.model_output = output
        return audio

    def run(self, audio=None):
        if audio is None:
            audio = self.forward()
        self.audio = audio
        return self.audio

    def make_audio_signals(self):
        estimates = []
        for i in range(self.audio.shape[-1]):
            _estimate = self.audio_signal.make_copy_with_audio_data(
                self.audio[..., i])
            estimates.append(_estimate)
        return estimates



"""
Robust principle component analysis
-----------------------------------

.. autoclass:: nussl.separation.factorization.RPCA
    :autosummary:

Independent component analysis
------------------------------

.. autoclass:: nussl.separation.factorization.ICA
    :autosummary:

"""

from .rpca import RPCA
from .ica import ICA



import numpy as np

from .. import MaskSeparationBase
from ..benchmark import HighLowPassFilter
from ...core import utils
from ...core import constants


class RepetSim(MaskSeparationBase):
    """
    Implements the REpeating Pattern Extraction Technique algorithm using 
    the Similarity Matrix (REPET-SIM).

    REPET is a simple method for separating the repeating background from the 
    non-repeating foreground in a piece of audio mixture. REPET-SIM is a generalization 
    of REPET, which looks for similarities instead of periodicities.

    References:

    [1] Zafar Rafii and Bryan Pardo. 
        "Music/Voice Separation using the Similarity Matrix," 
        13th International Society on Music Information Retrieval, 
        Porto, Portugal, October 8-12, 2012.

    Args:
        input_audio_signal (AudioSignal): Audio signal to be separated.
        similarity_threshold (int, optional): Threshold for considering two
          frames to be similar. Defaults to 0.
        min_distance_between_frames (float, optional): Number of seconds two frames
          must be apart to be considered neighbors. Defaults to 1.
        max_repeating_frames (int, optional): Max number of frames to consider as
          neighbors. Defaults to 100.
        high_pass_cutoff (float, optional): Cutoff for high pass filters. Bins below
          this cutoff will be given to the background. Defaults to 100.
        mask_type (str, optional): Mask type to use.. Defaults to 'soft'.
        mask_threshold (float, optional): Threshold for mask converting to binary. 
          Defaults to 0.5.
    """

    def __init__(self, input_audio_signal, similarity_threshold=0, 
                 min_distance_between_frames=1, max_repeating_frames=100, 
                 high_pass_cutoff=100, mask_type='soft', mask_threshold=0.5):
        super().__init__(
            input_audio_signal=input_audio_signal, 
            mask_type=mask_type,
            mask_threshold=mask_threshold)

        self.high_pass_cutoff = high_pass_cutoff
        self.similarity_threshold = similarity_threshold
        self.min_distance_between_frames = min_distance_between_frames
        self.max_repeating_frames = max_repeating_frames

        self._min_distance_converted_to_hops = False

        self.similarity_matrix = None
        self.similarity_indices = None
        self.magnitude_spectrogram = None

    def run(self):
        high_low = HighLowPassFilter(self.audio_signal, self.high_pass_cutoff)
        high_pass_masks = high_low.run()

        self.magnitude_spectrogram = np.abs(self.stft)

        background_masks = []
        foreground_masks = []

        self.similarity_indices = self._get_similarity_indices()

        for ch in range(self.audio_signal.num_channels):
            background_mask = self._compute_mask(
                self.magnitude_spectrogram[..., ch])
            foreground_mask = 1 - background_mask
            
            background_masks.append(background_mask)
            foreground_masks.append(foreground_mask)

        background_masks = np.stack(background_masks, axis=-1)
        foreground_masks = np.stack(foreground_masks, axis=-1)

        _masks = np.stack([background_masks, foreground_masks], axis=-1)

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)
            
            if i == 0:
                mask_data = np.maximum(mask_data, high_pass_masks[i].mask)
            elif i == 1:
                mask_data = np.minimum(mask_data, high_pass_masks[i].mask)
            
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks

    def _get_similarity_indices(self):
        self.similarity_matrix = self.get_similarity_matrix()

        if not self._min_distance_converted_to_hops:
            self.min_distance_between_frames *= (
                self.audio_signal.sample_rate / self.stft_params.hop_length
            )
            self._min_distance_converted_to_hops = True

        return self._find_similarity_indices()

    @staticmethod
    def compute_similarity_matrix(matrix):
        """Computes the cosine similarity matrix using the cosine similarity for any given input matrix X.

        Parameters:
            matrix (np.array): 2D matrix containing the magnitude spectrogram of the audio signal
        Returns:
            S (np.array): 2D similarity matrix
        """
        # ignore the 'divide by zero' warning
        with np.errstate(divide='ignore'):
            # Code below is adopted from http://stackoverflow.com/a/20687984/5768001

            # base similarity matrix (all dot products)
            similarity = np.dot(matrix, matrix.T)

            # inverse_mask of the squared magnitude of preference vectors (number of occurrences)
            inv_square_mag = 1 / np.diag(similarity)

            # if it doesn't occur, set it's inverse_mask magnitude to zero (instead of inf)
            inv_square_mag[np.isinf(inv_square_mag)] = 0

            # inverse_mask of the magnitude
            inv_mag = np.sqrt(inv_square_mag)

            # cosine similarity (element-wise multiply by inverse_mask magnitudes)
            cosine = similarity * inv_mag
            cosine = cosine.T * inv_mag

            return cosine

    def _find_similarity_indices(self):
        """Finds the similarity indices for all time frames from the similarity matrix

        Returns:
            similarity_indices (list of lists): similarity indices for all time frames
        """
        similarity_indices = []
        for i in range(self.audio_signal.stft_length):
            cur_indices = utils.find_peak_indices(
                self.similarity_matrix[i, :], self.max_repeating_frames,
                min_dist=self.min_distance_between_frames,
                threshold=self.similarity_threshold)

            # the first peak is always itself so we throw it out
            # we also want only self.max_repeating_frames peaks
            # so +1 for 0-based, and +1 for the first peak we threw out
            cur_indices = cur_indices[1:self.max_repeating_frames + 2]
            similarity_indices.append(cur_indices)

        return similarity_indices

    def _compute_mask(self, magnitude_spectrogram_channel):
        mask = np.zeros_like(magnitude_spectrogram_channel)

        for i in range(self.audio_signal.stft_length):
            cur_similarities = self.similarity_indices[i]

            if not cur_similarities:
                # If there are no similarities, then just add ones to the mask here.
                mask[:, i] = np.ones(mask.shape[constants.STFT_VERT_INDEX])
            else:
                similar_times = np.array([magnitude_spectrogram_channel[:, j]
                                          for j in cur_similarities])

                mask[:, i] = np.median(similar_times, axis=0)

        mask = np.minimum(mask, magnitude_spectrogram_channel)
        mask = (mask + constants.EPSILON) / (magnitude_spectrogram_channel + constants.EPSILON)
        return mask

    def get_similarity_matrix(self):
        """Calculates and returns the similarity matrix for the audio file associated with this object

        Returns:
             similarity_matrix (np.array): similarity matrix for the audio file.

        EXAMPLE::

            # Set up audio signal
            signal = nussl.AudioSignal('path_to_file.wav')

            # Set up a RepetSim object
            repet_sim = nussl.RepetSim(signal)

            # I don't have to run repet to get a similarity matrix for signal
            sim_mat = repet_sim.get_similarity_matrix()

        """
        mean_magnitude_spectrogram = np.mean(self.magnitude_spectrogram, axis=2)
        return self.compute_similarity_matrix(mean_magnitude_spectrogram.T)



from ..base import MaskSeparationBase, SeparationException
from ...datasets import transforms


class IdealRatioMask(MaskSeparationBase):
    """
    Implements an ideal ratio mask (IRM) that is computed by using the known
    ground truth performance. This is one of the upper baselines.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        sources (list): List of audio signal objects that correspond to the sources.
        approach (str): Either 'psa' (phase sensitive spectrum approximation) or 'msa'
          (magnitude spectrum approximation). Generally 'psa' does better.
        mask_type (str, optional): Mask type. Defaults to 'soft'.
        mask_threshold (float, optional): Masking threshold. Defaults to 0.5. 
        kwargs (dict): Extra keyword arguments are passed to the transform classes at
          initialization.
    """

    def __init__(self, input_audio_signal, sources, approach='psa', mask_type='soft',
                 mask_threshold=.5, **kwargs):

        if isinstance(sources, list):
            sources = {i: sources[i] for i in range(len(sources))}
        elif not isinstance(sources, dict):
            raise SeparationException("sources must be a list or a dict!")

        self.sources = sources
        self.approach = approach
        if self.approach == 'psa':
            tfm = transforms.PhaseSensitiveSpectrumApproximation(**kwargs)
        elif self.approach == 'msa':
            tfm = transforms.MagnitudeSpectrumApproximation(**kwargs)
        else:
            raise SeparationException(f'Unknown approach: {self.approach}')
        self.tfm = tfm

        super().__init__(
            input_audio_signal=input_audio_signal, mask_type=mask_type,
            mask_threshold=mask_threshold)

    def run(self):
        # Set up dictionary to pass to the transform.    
        data = {
            'mix': self.audio_signal,
            'sources': self.sources
        }
            
        output = self.tfm(data)
        masks = []
        mask_data = (
            output['source_magnitudes'] / 
            (output['source_magnitudes'].sum(axis=-1, keepdims=True) + 1e-8)
        )

        for i in range(mask_data.shape[-1]):            
            mask = self.mask_type(mask_data[..., i])
            masks.append(mask)

        self.result_masks = masks
        return self.result_masks



import pytest
from nussl import efz_utils
import tempfile
import os
import musdb
import zipfile
import scaper
import random
import glob
import nussl
from nussl.datasets import transforms
from nussl import datasets
import numpy as np
import torch
import json


def _unzip(path_to_zip, target_path):
    with zipfile.ZipFile(path_to_zip, 'r') as zip_ref:
        zip_ref.extractall(target_path)


fix_dir = os.path.expanduser('~/.nussl/tests/')
os.makedirs(fix_dir, exist_ok=True)
OVERWRITE_REGRESSION_DATA = False


@pytest.fixture(scope="module")
def benchmark_audio():
    audio_files = {}
    keys = ['K0140.wav', 'K0149.wav', 'dev1_female3_inst_mix.wav']
    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        for k in keys:
            audio_files[k] = efz_utils.download_audio_file(k, _dir)
        yield audio_files


@pytest.fixture(scope="module")
def musdb_tracks():
    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        db = musdb.DB(root=_dir, download=True)
        yield db


@pytest.fixture(scope="module")
def toy_datasets():
    dataset_locations = {}
    keys = ['babywsj_oW0F0H9.zip']
    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        for k in keys:
            target_folder = os.path.join(_dir, os.path.splitext(k)[0])
            data = efz_utils.download_benchmark_file(k, _dir)
            _unzip(data, target_folder)
            dataset_locations[k] = target_folder
        yield dataset_locations


@pytest.fixture(scope="module")
def mix_source_folder(toy_datasets):
    wsj_sources = toy_datasets['babywsj_oW0F0H9.zip']
    audio_files = glob.glob(
        f"{wsj_sources}/**/*.wav", recursive=True)
    n_sources = 2
    n_mixtures = 10

    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        _dir = os.path.join(_dir, 'mix_source_folder')
        for i in range(n_mixtures):
            sources = []
            for n in range(n_sources):
                path = random.choice(audio_files)
                source = nussl.AudioSignal(path)
                sources.append(source)

            min_length = min([s.signal_length for s in sources])

            for n in range(n_sources):
                output_path = os.path.join(_dir, f's{n}', f'{i}.wav')
                os.makedirs(os.path.dirname(output_path), exist_ok=True)
                sources[n].truncate_samples(min_length)
                sources[n].write_audio_to_file(output_path)

            mix = sum(sources)
            output_path = os.path.join(_dir, 'mix', f'{i}.wav')
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            mix.write_audio_to_file(output_path)
        yield _dir


@pytest.fixture(scope="module")
def scaper_folder(toy_datasets):
    wsj_sources = toy_datasets['babywsj_oW0F0H9.zip']
    fg_path = os.path.join(
        wsj_sources, 'babywsj', 'dev')

    n_sources = 2
    n_mixtures = 10
    duration = 3
    ref_db = -40

    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        _dir = os.path.join(_dir, 'scaper')
        os.makedirs(_dir, exist_ok=True)
        for i in range(n_mixtures):
            sc = scaper.Scaper(
                duration, fg_path, fg_path, random_state=i)
            sc.ref_db = ref_db
            sc.sr = 16000

            for j in range(n_sources):
                sc.add_event(
                    label=('choose', []),
                    source_file=('choose', []),
                    source_time=('const', 0),
                    event_time=('const', 0),
                    event_duration=('const', duration),
                    snr=('const', 0),
                    pitch_shift=None,
                    time_stretch=None
                )

            audio_path = os.path.join(_dir, f'{i}.wav')
            jams_path = os.path.join(_dir, f'{i}.jams')
            sc.generate(audio_path, jams_path, save_isolated_events=True)

        yield _dir


@pytest.fixture(scope="module")
def mix_and_sources(scaper_folder):
    dataset = datasets.Scaper(scaper_folder)
    item = dataset[0]
    return item['mix'], item['sources']


@pytest.fixture(scope="module")
def music_mix_and_sources(musdb_tracks):
    dataset = datasets.MUSDB18(
        folder=musdb_tracks.root, download=False,
        transform=transforms.SumSources(
            [['drums', 'bass', 'other']]))
    item = dataset[0]
    return item['mix'], item['sources']


@pytest.fixture(scope="module")
def drum_and_vocals(musdb_tracks):
    dataset = datasets.MUSDB18(
        folder=musdb_tracks.root, download=False)
    item = dataset[0]
    return item['sources']['drums'], item['sources']['vocals']


@pytest.fixture(scope="module")
def bad_scaper_folder(toy_datasets):
    wsj_sources = toy_datasets['babywsj_oW0F0H9.zip']
    fg_path = os.path.join(
        wsj_sources, 'babywsj', 'dev')

    n_sources = 2
    n_mixtures = 10
    duration = 3
    ref_db = -40

    with tempfile.TemporaryDirectory() as tmp_dir:
        _dir = tmp_dir if fix_dir is None else fix_dir
        _dir = os.path.join(_dir, 'bad_scaper')
        os.makedirs(_dir, exist_ok=True)
        for i in range(n_mixtures):
            sc = scaper.Scaper(
                duration, fg_path, fg_path, random_state=i)
            sc.ref_db = ref_db
            sc.sr = 16000

            for j in range(n_sources):
                sc.add_event(
                    label=('choose', []),
                    source_file=('choose', []),
                    source_time=('const', 0),
                    event_time=('const', 0),
                    event_duration=('const', duration),
                    snr=('const', 0),
                    pitch_shift=None,
                    time_stretch=None
                )

            audio_path = os.path.join(_dir, f'{i}.wav')
            jams_path = os.path.join(_dir, f'{i}.jams')
            sc.generate(audio_path, jams_path, save_isolated_events=False)

        yield _dir


@pytest.fixture(scope="module")
def one_item(scaper_folder):
    stft_params = nussl.STFTParams(
        window_length=512,
        hop_length=128
    )
    tfms = transforms.Compose([
        transforms.PhaseSensitiveSpectrumApproximation(),
        transforms.GetAudio(),
        transforms.ToSeparationModel()
    ])
    dataset = nussl.datasets.Scaper(
        scaper_folder, transform=tfms, stft_params=stft_params)
    i = np.random.randint(len(dataset))
    data = dataset[i]
    for k in data:
        # fake a batch dimension
        if torch.is_tensor(data[k]):
            data[k] = data[k].unsqueeze(0)
    yield data


@pytest.fixture(scope="module")
def check_against_regression_data():
    def check(scores, path):
        if not os.path.exists(path):
            with open(path, 'w') as f:
                json.dump(scores, f, indent=4)
        if OVERWRITE_REGRESSION_DATA:
            with open(path, 'w') as f:
                json.dump(scores, f, indent=4)
        else:
            with open(path, 'r') as f:
                reg_scores = json.load(f)
            for key in scores:
                if key not in ['permutation', 'combination']:
                    for metric in scores[key]:
                        if metric in reg_scores[key]:
                            assert np.allclose(
                                scores[key][metric],
                                reg_scores[key][metric],
                                atol=1e-1
                            )
    return check




"""
Base for all methods
--------------------

.. autoclass:: nussl.separation.SeparationBase
    :members:
    :autosummary:

Base for masking-based methods
------------------------------

.. autoclass:: nussl.separation.MaskSeparationBase
    :members:
    :autosummary:

Base for clustering-based methods
---------------------------------

.. autoclass:: nussl.separation.ClusteringSeparationBase
    :members:
    :autosummary:

Mix-in for NMF-based methods
----------------------------

.. autoclass:: nussl.separation.NMFMixin
    :members:
    :autosummary:

Mix-in for deep methods
-----------------------

.. autoclass:: nussl.separation.DeepMixin
    :members:
    :autosummary:

"""

from .separation_base import SeparationBase, SeparationException
from .mask_separation_base import MaskSeparationBase
from .clustering_separation_base import ClusteringSeparationBase
from .deep_mixin import DeepMixin
from .nmf_mixin import NMFMixin



import numpy as np
import librosa

from .. import MaskSeparationBase


class HPSS(MaskSeparationBase):
    """
    Implements harmonic/percussive source separation based on [1]. This is a 
    wrapper around the librosa implementation.

    References:
    
    [1] Fitzgerald, Derry. “Harmonic/percussive separation using median filtering.” 
        13th International Conference on Digital Audio Effects (DAFX10), Graz, 
        Austria, 2010.
    
    [2] Driedger, Müller, Disch. “Extending harmonic-percussive separation of audio.” 
        15th International Society for Music Information Retrieval Conference 
        (ISMIR 2014) Taipei, Taiwan, 2014.

    Parameters:
        input_audio_signal (AudioSignal): signal to separate.

        kernel_size (int or tuple (kernel_harmonic, kernel_percussive)): 
          kernel size(s) for the median filters.

        mask_type (str, optional): Mask type. Defaults to 'soft'.

        mask_threshold (float, optional): Masking threshold. Defaults to 0.5.

    """
    def __init__(self, input_audio_signal, kernel_size=31, mask_type='soft',
                 mask_threshold=0.5):
        super().__init__(
            input_audio_signal=input_audio_signal, 
            mask_type=mask_type,
            mask_threshold=mask_threshold
        )

        self.kernel_size = kernel_size

    def run(self):
        # separate the mixture background by masking
        harmonic_masks = []
        percussive_masks = []

        for ch in range(self.audio_signal.num_channels):
            # apply mask
            harmonic_mask, percussive_mask = librosa.decompose.hpss(
                self.stft[:, :, ch], kernel_size=self.kernel_size, mask=True)
            harmonic_masks.append(harmonic_mask)
            percussive_masks.append(percussive_mask)

        # make a mask and return
        harmonic_masks = np.stack(harmonic_masks, axis=-1)
        percussive_masks = np.stack(percussive_masks, axis=-1)

        _masks = np.stack([harmonic_masks, percussive_masks], axis=-1)
        
        self.result_masks = []
        
        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks



"""
Small collection of utilities for altering and remixing
AudioSignal objects. 
"""
import copy

import numpy as np

from . import AudioSignal


def pan_audio_signal(audio_signal, angle_in_degrees):
    """
    Pans an audio signal left or right by the desired number of degrees. This
    returns a copy of the input audio signal.

    Use negative numbers to pan left, positive to pan right. Angles outside of the
    range [-45, 45] raise an error.
    
    Args:
        audio_signal (AudioSignal): Audio signal to be panned.
        angle_in_degrees (float): Angle in degrees to pan by, between -45 and 45.
    
    Raises:
        ValueError: Angles outside of the range [-45, 45] raise an error.
    
    Returns:
        AudioSignal: Audio signal panned by `angle_in_degrees`.
    """
    if angle_in_degrees < -45 or angle_in_degrees > 45:
        raise ValueError(
            "Angle must be between -45 and 45! -45 means "
            "all the way to the left, 45 means all the way "
            "to the right, and 0 is center.")
    panned_signal = copy.deepcopy(audio_signal)
    panned_signal.to_mono()

    panned_signal.audio_data = np.concatenate([
        panned_signal.audio_data, panned_signal.audio_data
    ], axis=0)

    angle = np.deg2rad(angle_in_degrees)
    left_scale = (
        np.sqrt(2)/2 * (np.cos(angle) - np.sin(angle)))
    right_scale = (
        np.sqrt(2)/2 * (np.cos(angle) + np.sin(angle)))
    
    panned_signal.audio_data[0] *= left_scale
    panned_signal.audio_data[1] *= right_scale

    return panned_signal


def delay_audio_signal(audio_signal, delays_in_samples):
    """
    Delays an audio signal by the desired number of samples per channel. This
    returns a copy of the input audio signal.

    Delay must be positive. The end of the audio signal is truncated for that
    channel so that the length remains the same as the original.
    
    Args:
        audio_signal (AudioSignal): Audio signal to be panned.
        delays_in_samples (list of int): List of delays to apply to each channel.
          Should have the same length as number of channels in the AudioSignal.
    
    Raises:
        ValueError: If length of `delays_in_samples`does not match number of channels
          in the audio signal. Or if any items in `delays_in_samples` are float type. Or
          if any delays are negative.
    
    Returns:
        AudioSignal: Audio signal with each channel delayed by the specified number
          samples in `delays_in_samples`.
    """
    if any([not isinstance(x, int) for x in delays_in_samples]):
        raise ValueError("All items in delay_in_samples must be integers.")
    if any([x < 0 for x in delays_in_samples]):
        raise ValueError("All items in delay_in_samples must be non-negative.")
    if len(delays_in_samples) != audio_signal.num_channels:
        raise ValueError(
            "Number of items in delays_in_samples must match number of "
            "channels in audio_signal.")
            
    delayed_signal = copy.deepcopy(audio_signal)

    for i, delay in enumerate(delays_in_samples):
        if delay > 0:
            _audio_data = delayed_signal.audio_data[i]
            original_length = _audio_data.shape[-1]

            _audio_data = np.pad(_audio_data, (delay, 0))
            _audio_data = _audio_data[:original_length]

            delayed_signal.audio_data[i] = _audio_data
    
    return delayed_signal



"""
Base class for Mask objects. Contains many common utilities used for accessing masks. The mask itself is
represented under the hood as a three dimensional numpy :obj:`ndarray` object. The dimensions are 
``[NUM_FREQ, NUM_HOPS, NUM_CHAN]``. Safe accessors for these array indices are in :ref:`constants` as well as 
below.

Right now only spectrogram-like masks are supported (note the shape of the :ref:`mask` property), but in future
releases nussl will support masks for representations with different dimensionality requirements.
"""

import numbers

import numpy as np

from .. import utils
from .. import constants


class MaskBase(object):
    """
    Args:
        input_mask (:obj:`np.ndarray`): A 2- or 3-dimensional numpy ``ndarray`` representing a mask.
        
    """
    def __init__(self, input_mask=None, mask_shape=None):
        self._mask = None

        if mask_shape is None and input_mask is None:
            raise ValueError('Cannot initialize mask without mask_shape or input_mask!')
        if mask_shape is not None and input_mask is not None:
            raise ValueError('Cannot initialize mask with both mask_shape and input_mask!')
        if isinstance(input_mask, np.ndarray):
            self.mask = input_mask
        elif isinstance(mask_shape, tuple):
            self.mask = np.zeros(mask_shape)
        else:
            raise ValueError('input_mask must be a np.ndarray, or mask_shape must be a tuple!')

    @property
    def mask(self):
        """        
        The actual mask. This is represented as a three dimensional numpy :obj:`ndarray` object.
        The input gets validated by :func:`_validate_mask`. In the case of 
        :class:`separation.masks.binary_mask.BinaryMask` the validation checks that the values are all 1 or 0 
        (or bools), in the case of :class:`separation.masks.soft_mask.SoftMask` the validation checks that all values
        are within the domain ``[0.0, 1.0]``.
        
        This base class will throw a ``NotImplementedError`` if instantiated directly.
        
        Raises:
            :obj:`ValueError` if :attr:`mask.ndim` is less than 2 or greater than 3, or if values fail validation.
            :obj:`NotImplementedError` if instantiated directly.

        """
        return self._mask

    @mask.setter
    def mask(self, value):
        assert isinstance(value, np.ndarray), 'Type of self.mask must be np.ndarray!'

        if value.ndim == 1:
            raise ValueError('Cannot support arrays with less than 2 dimensions!')

        if value.ndim == 2:
            value = np.expand_dims(value, axis=constants.STFT_CHAN_INDEX)

        if value.ndim > 3:
            raise ValueError('Cannot support arrays with more than 3 dimensions!')

        self._mask = self._validate_mask(value)

    def get_channel(self, ch):
        """
        Gets mask channel ``ch`` and returns it as a 2D :obj:`np.ndarray`
        
        Args:
            ch (int): Channel index to return (0-based).

        Returns:
            :obj:`np.array` with the mask channel
            
        Raises:
            :obj:`ValueError` if ``ch`` is less than 0 or greater than the number of channels that this mask object has.

        """
        if ch >= self.num_channels:
            raise ValueError(
                f'Cannot get channel {ch} for object w/ {self.num_channels} channels!'
                ' (0-based)'
            )

        if ch < 0:
            raise ValueError(f'Cannot get channel {ch}. This will cause unexpected results!')

        return utils._get_axis(self.mask, constants.STFT_CHAN_INDEX, ch)

    @property
    def num_channels(self):
        """
        (int) Number of channels this mask has.

        """
        return self.mask.shape[constants.STFT_CHAN_INDEX]

    @property
    def shape(self):
        """
        (tuple) Returns the shape of the whole mask. Identical to ``np.ndarray.shape()``.

        """
        return self.mask.shape

    @property
    def dtype(self):
        """
        (str) Returns the data type of the values of the mask. 

        """
        return self.mask.dtype

    @staticmethod
    def _validate_mask(mask_):
        """
        Args:
            mask_: 

        Returns:

        """
        raise NotImplementedError('Cannot call base class! Use BinaryMask or SoftMask!')

    @classmethod
    def ones(cls, shape):
        """
        Makes a mask with all ones with the specified shape. Exactly the same as ``np.ones()``.
        Args:
            shape (tuple): Shape of the resultant mask.

        Returns:

        """
        return cls(np.ones(shape))

    @classmethod
    def zeros(cls, shape):
        """
        Makes a mask with all zeros with the specified shape. Exactly the same as ``np.zeros()``.
        Args:
            shape (tuple): Shape of the resultant mask.

        Returns:

        """
        return cls(np.zeros(shape))

    def invert_mask(self):
        """

        Returns:

        """
        raise NotImplementedError('Cannot call base class! Use BinaryMask or SoftMask!')

    def inverse_mask(self):
        """
        Alias for :func:`invert_mask`

        See Also:
            :func:`invert_mask`

        Returns:

        """
        return self.invert_mask()

    def _add(self, other):
        class_method = type(self)
        new_mask = class_method(self.mask)
        if isinstance(other, MaskBase):
            new_mask.mask = new_mask.mask + other.mask
            return new_mask
        if isinstance(other, np.ndarray):
            new_mask.mask = new_mask.mask + other
            return new_mask
        else:
            raise ValueError(f'Cannot do arithmetic operation with MaskBase and {type(other)}')

    def _mult(self, value):
        if not isinstance(value, numbers.Real):
            raise ValueError(f'Cannot do operation with MaskBase and {type(value)}')
        class_method = type(self)
        new_mask = class_method(self.mask)
        new_mask.mask = new_mask.mask * value
        return new_mask

    def __add__(self, other):
        return self._add(other)

    def __sub__(self, other):
        return self + (-1 * other)

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __mul__(self, value):
        return self._mult(value)

    def __rmul__(self, value):
        return self._mult(value)

    def __div__(self, value):
        return self._mult(1 / float(value))

    def __truediv__(self, value):
        return self.__div__(value)

    def __imul__(self, value):
        return self * value

    def __idiv__(self, value):
        return self / value

    def __itruediv__(self, value):
        return self.__idiv__(value)

    def __eq__(self, other):
        return np.array_equal(self.mask, other.mask)

    def __ne__(self, other):
        return not self.__eq__(other)



"""
The :class:`BinaryMask` class is for creating a time-frequency mask with binary values. Like all 
:class:`separation.masks.mask_base.MaskBase` objects, :class:`BinaryMask` is initialized with a 2D or 3D numpy array
containing the mask data. The data type (numpy.dtype) of the initial mask can be either bool, int, or float. 
The mask is stored as a 3-dimensional boolean-valued numpy array.

The best case scenario for the input mask np array is when the data type is bool. If the data type of the input mask 
upon init is int it is expected that all values are either 0 or 1. If the data type
of the mask is float, all values must be within 1e-2 of either 1 or 0. If the array is not set as one of these, 
:class:`BinaryMask` will raise an exception.

:class:`BinaryMask` (like :class:`separation.masks.soft_mask.SoftMask`) is one of the return types for the :func:`run()` 
methods of :class:`separation.mask_separation_base.MaskSeparationBase`-derived objects (this is most of the 
separation methods in `nussl`.

See Also:
    * :class:`separation.masks.mask_base.MaskBase`: The base class for BinaryMask and SoftMask
    * :class:`separation.masks.soft_mask.SoftMask`: Similar to BinaryMask, but instead of taking boolean values, 
      takes floats between ``[0.0 and 1.0]``.
    * :class:`separation.mask_separation_base.MaskSeparationBase`: Base class for all mask-based separation methods 
      in `nussl`.

Examples:
    Initializing a mask from a numpy array...
    
.. code-block:: python
    :linenos:
    
    import nussl
    import numpy as np
    
    # load a file
    signal = nussl.AudioSignal('path/to/file.wav')
    stft = signal.stft()
    
    # Make a random binary mask with the same shape as the stft with dtype == bool
    rand_bool_mask = np.random.randint(2, size=stft.shape).astype('bool')
    bin_mask_bool = nussl.BinaryMask(rand_bool_mask)
    
    # Make a random binary mask with the same shape as the stft with dtype == int
    rand_int_mask = np.random.randint(2, size=stft.shape)
    bin_mask_int = nussl.BinaryMask(rand_int_mask)
    
    # Make a random binary mask with the same shape as the stft with dtype == float
    rand_float_mask = np.random.randint(2, size=stft.shape).astype('float')
    bin_mask_int = nussl.BinaryMask(rand_float_mask)    
    
    
:class:`separation.mask_separation_base.MaskSeparationBase`-derived methods return 
:class:`separation.masks.mask_base.MaskBase` masks, like so...

.. code-block:: python
    :linenos:

    import nussl
    
    # load a file
    signal = nussl.AudioSignal('path/to/file.wav')
    
    repet = nussl.Repet(signal, mask_type=nussl.BinaryMask)  # You have to specify that you want Binary Masks back
    assert isinstance(repet, nussl.MaskSeparationBase)  # Repet is a MaskSeparationBase-derived class
    
    [background_mask, foreground_mask] = repet.run()  # MaskSeparationBase-derived classes return MaskBase objects
    assert isinstance(foreground_mask, nussl.BinaryMask)  # this is True
    assert isinstance(background_mask, nussl.BinaryMask)  # this is True

"""

import numpy as np

from . import MaskBase


class BinaryMask(MaskBase):
    """
    Class for creating a Binary Mask to apply to a time-frequency representation of 
    the audio. 
    
    Args:
        input_mask (:obj:`np.ndarray`): 2- or 3-D :obj:`np.array` that represents the mask.
    """

    def __init__(self, input_mask=None, mask_shape=None):
        super(BinaryMask, self).__init__(input_mask, mask_shape)

    @staticmethod
    def _validate_mask(mask_):
        assert isinstance(mask_, np.ndarray), 'Mask must be a numpy array!'

        if mask_.dtype == np.bool:
            # This is perfect, do nothing here
            return mask_
        elif mask_.dtype.kind in np.typecodes['AllInteger']:
            if np.max(mask_) > 1 or np.min(mask_) < 0:
                raise ValueError('Found values in mask that are not 0 or 1. Mask must be binary!')
        elif mask_.dtype.kind in np.typecodes['AllFloat']:
            tol = 1e-2
            # If we have a float array, ensure that all values are close to 1 or 0
            if not np.all(np.logical_or(np.isclose(mask_, [0], atol=tol), np.isclose(mask_, [1], atol=tol))):
                raise ValueError('All mask values must be close to 0 or 1!')

        return mask_.astype('bool')

    def mask_as_ints(self, channel=None):
        """
        Returns this :class:`BinaryMask` as a numpy array of ints of 0's and 1's.
        
        Returns:
            numpy :obj:`ndarray` of this :obj:`BinaryMask` represented as ints instead of bools.

        """
        if channel is None:
            return self.mask.astype(int)
        else:
            return self.get_channel(channel).astype(int)

    def invert_mask(self):
        """
        Makes a new :class:`BinaryMask` object with a logical not applied to flip the values in this :class:`BinaryMask`
        object.

        Returns:
            A new :class:`BinaryMask` object that has all of the boolean values flipped.

        """
        return BinaryMask(np.logical_not(self.mask))



import os
import logging
import copy
import time
from datetime import timedelta

from ignite.engine import Events, Engine, EventEnum
from ignite.handlers import Timer
from ignite.contrib.handlers import ProgressBar
from ignite.metrics import RunningAverage
from torch.utils.tensorboard import SummaryWriter
import torch
from torch import nn
import numpy as np

from nussl import datasets


class ValidationEvents(EventEnum):
    """
    Events based on validation running
    """
    VALIDATION_STARTED = 'validation_started'
    VALIDATION_COMPLETED = 'validation_completed'


class BackwardsEvents(EventEnum):
    """
    Events based on validation running
    """
    BACKWARDS_COMPLETED = 'backwards_completed'


def cache_dataset(dataset):
    """
    Runs through an entire dataset and caches it if there nussl.datasets.transforms.Cache
    is in dataset.transform. If there is no caching, or dataset.cache_populated = True,
    then this function just iterates through the dataset and does nothing.

    This function can also take a `torch.util.data.DataLoader` object wrapped around
    a `nussl.datasets.BaseDataset` object.
    
    Args:
        dataset (nussl.datasets.BaseDataset): Must be a subclass of 
          `nussl.datasets.BaseDataset`.
    """

    def dummy_process(engine, data):
        pass

    cache = Engine(dummy_process)
    ProgressBar().attach(cache)
    cache.run(dataset)
    dataset.cache_populated = True


def create_train_and_validation_engines(train_func, val_func=None, device='cpu'):
    """
    Helper function for creating an ignite Engine object with helpful defaults.
    This sets up an Engine that has four handlers attached to it:

    - prepare_batch: before a batch is passed to train_func or val_func, this
      function runs, moving every item in the batch (which is a dictionary) to
      the appropriate device ('cpu'  or 'cuda').

    - book_keeping: sets up some dictionaries that are used for bookkeeping so one
      can easily track the epoch and iteration losses for both training and
      validation.

    - add_to_iter_history: records the iteration, epoch, and past iteration losses
      into the dictionaries set up by book_keeping.

    - clear_iter_history: resets the current iteration history of losses after moving
      the current iteration history into past iteration history.
    
    Args:
        train_func (func): Function that provides the closure for training for
          a single batch.
        val_func (func, optional): Function that provides the closure for
          validating a single batch. Defaults to None.
        device (str, optional): Device to move tensors to. Defaults to 'cpu'.
    """
    # Set up engines for training and validation
    trainer = Engine(train_func)
    trainer.register_events(*ValidationEvents)
    trainer.register_events(*BackwardsEvents)

    validator = None if val_func is None else Engine(val_func)

    # Before a batch starts, the items should be float and moved to the 
    # correct device, for both training and validation. Checks to make
    # sure "cuda" is available if user requested cuda.
    device = device if torch.cuda.is_available() else 'cpu'
    device = torch.device(device)

    def prepare_batch(engine):
        batch = engine.state.batch
        for key in batch:
            if torch.is_tensor(batch[key]):
                batch[key] = batch[key].float().to(device)
        engine.state.batch = batch

    # Set up stuff for bookkeeping as training progresses.
    def book_keeping(engine):
        engine.state.epoch_history = {}
        engine.state.iter_history = {}
        engine.state.past_iter_history = {}

    def add_to_iter_history(engine):
        for key in engine.state.output:
            if key not in engine.state.iter_history:
                engine.state.iter_history[key] = []
            if key not in engine.state.past_iter_history:
                engine.state.past_iter_history[key] = []
            engine.state.iter_history[key].append(
                engine.state.output[key]
            )
            engine.state.past_iter_history[key].append(
                engine.state.iter_history[key]
            )

    def clear_iter_history(engine):
        engine.state.iter_history = {}

    trainer.add_event_handler(
        Events.ITERATION_STARTED, prepare_batch)
    trainer.add_event_handler(
        Events.STARTED, book_keeping)
    trainer.add_event_handler(
        Events.ITERATION_COMPLETED, add_to_iter_history)
    trainer.add_event_handler(
        Events.EPOCH_STARTED, clear_iter_history)

    if validator is not None:
        validator.add_event_handler(
            Events.ITERATION_STARTED, prepare_batch)
        validator.add_event_handler(
            Events.STARTED, book_keeping)
        validator.add_event_handler(
            Events.ITERATION_COMPLETED, add_to_iter_history)
        validator.add_event_handler(
            Events.EPOCH_STARTED, clear_iter_history)

    return trainer, validator


def _remove_cache_from_tfms(transforms):
    transforms = copy.deepcopy(transforms)

    if isinstance(transforms, datasets.transforms.Compose):
        for t in transforms.transforms:
            if isinstance(t, datasets.transforms.Cache):
                transforms.transforms.remove(t)

    return transforms


def _prep_metadata(metadata):
    metadata = copy.deepcopy(metadata)
    metadata['transforms'] = _remove_cache_from_tfms(metadata['transforms'])
    return metadata


def add_validate_and_checkpoint(output_folder, model, optimizer, train_data, trainer,
                                val_data=None, validator=None, save_by_epoch=None):
    """
    This adds the following handler to the trainer:

    - validate_and_checkpoint: this runs the validator on the validation dataset 
      (``val_data``) using a defined validation process function ``val_func``. 
      These are optional. If these are not provided, then no validator is run
      and the model is simply checkpointed. The model is always saved to 
      ``{output_folder}/checkpoints/latest.model.pth``. If the model is also the 
      one with the lowest validation loss, then it is *also* saved to
      ``{output_folder}/checkpoints/best.model.pth. This is attached to
      ``Events.EPOCH_COMPLETED`` on the trainer. After completion, it fires a
      ``ValidationEvents.VALIDATION_COMPLETED`` event.

    Args:
        model (torch.nn.Module): Model that is being trained (typically a SeparationModel).
          optimizer (torch.optim.Optimizer): Optimizer being used to train.

        train_data (BaseDataset): dataset that is being used to train the model. This is to
          save additional metadata information alongside the model checkpoint such as the
          STFTParams, dataset folder, length, list of transforms, etc.

        trainer (ignite.Engine): Engine for trainer

        validator (ignite.Engine, optional): Engine for validation. 
          Defaults to None.

        val_data (torch.utils.data.Dataset, optional): The validation data. 
          Defaults to None.

        save_by_epoch (int, optional): Save the model by epoch number. If this is set to
          N, then every Nth model will be saved in the format epoch{N}.model.pth.
    """

    # When the trainer finishes an epoch, it should validate and save 
    # the model.
    @trainer.on(Events.EPOCH_COMPLETED)
    def validate_and_checkpoint(trainer):
        trainer.fire_event(ValidationEvents.VALIDATION_STARTED)

        is_best = True
        if validator is not None:
            validator.run(val_data)

            for key in validator.state.iter_history:
                _key = f"validation/{key}"
                if _key not in trainer.state.epoch_history:
                    trainer.state.epoch_history[_key] = []
                trainer.state.epoch_history[_key].append(np.mean(
                    validator.state.iter_history[key]
                ))

            if 'validation/loss' in trainer.state.epoch_history:
                cur = trainer.state.epoch_history['validation/loss'][-1]
                is_best = cur == min(trainer.state.epoch_history['validation/loss'])

        for key in trainer.state.iter_history:
            _key = f"train/{key}"
            if _key not in trainer.state.epoch_history:
                trainer.state.epoch_history[_key] = []
            trainer.state.epoch_history[_key].append(np.mean(
                trainer.state.iter_history[key]
            ))

        output_paths = [os.path.join(
            output_folder, 'checkpoints', 'latest.model.pth')]
        if is_best:
            output_paths.append(os.path.join(
                output_folder, 'checkpoints', 'best.model.pth'
            ))

        metadata = {
            'stft_params': train_data.stft_params,
            'sample_rate': train_data.sample_rate,
            'num_channels': train_data.num_channels,
            'train_dataset': _prep_metadata(train_data.metadata),
            'trainer.state_dict': {
                'epoch': trainer.state.epoch,
                'epoch_length': trainer.state.epoch_length,
                'max_epochs': trainer.state.max_epochs,
                'output': trainer.state.output,
                'metrics': trainer.state.metrics,
                'seed': trainer.state.seed,
            },
            'trainer.state.epoch_history': trainer.state.epoch_history,
        }
        try:
            # Store metadata for validation set if it exists
            metadata['val_dataset'] = _prep_metadata(val_data.metadata)
        except:
            pass

        if isinstance(model, nn.DataParallel):
            _model = model.module
        else:
            _model = model

        _model.metadata.update(metadata)

        for _path in output_paths:
            os.makedirs(os.path.join(
                output_folder, 'checkpoints'), exist_ok=True)
            _model.save(_path)
            torch.save(optimizer.state_dict(),
                       _path.replace('model.pth', 'optimizer.pth'))
        
        if save_by_epoch is not None:
            if trainer.state.epoch % save_by_epoch == 0:
                _path = output_paths[0].replace('latest', f'epoch{trainer.state.epoch}')
                _model.save(_path)

        trainer.state.saved_model_path = output_paths[-1]
        trainer.state.output_folder = output_folder
        trainer.fire_event(ValidationEvents.VALIDATION_COMPLETED)


def add_stdout_handler(trainer, validator=None):
    """
    This adds the following handler to the trainer engine, and also sets up
    Timers:

    - log_epoch_to_stdout: This logs the results of a model after it has trained
      for a single epoch on both the training and validation set. The output typically
      looks like this:

      .. code-block:: none

            EPOCH SUMMARY
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            - Epoch number: 0010 / 0010
            - Training loss:   0.583591
            - Validation loss: 0.137209
            - Epoch took: 00:00:03
            - Time since start: 00:00:32
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            Saving to test.
            Output @ tests/local/trainer
    
    Args:
        trainer (ignite.Engine): Engine for trainer

        validator (ignite.Engine, optional): Engine for validation. 
          Defaults to None.
    """
    # Set up timers for overall time taken and each epoch
    overall_timer = Timer(average=False)
    overall_timer.attach(trainer,
                         start=Events.STARTED, pause=Events.COMPLETED)

    epoch_timer = Timer(average=False)
    epoch_timer.attach(
        trainer, start=Events.EPOCH_STARTED,
        pause=ValidationEvents.VALIDATION_COMPLETED
    )

    @trainer.on(ValidationEvents.VALIDATION_COMPLETED)
    def log_epoch_to_stdout(trainer):
        epoch_time = epoch_timer.value()
        epoch_time = timedelta(seconds=epoch_time)
        overall_time = overall_timer.value()
        overall_time = timedelta(seconds=overall_time)

        epoch_number = trainer.state.epoch
        total_epochs = trainer.state.max_epochs

        try:
            validation_loss = (
                f"{trainer.state.epoch_history['validation/loss'][-1]:04f}")
        except:
            validation_loss = 'N/A'

        train_loss = trainer.state.epoch_history['train/loss'][-1]
        saved_model_path = trainer.state.saved_model_path

        logging_str = (
            f"\n\n"
            f"EPOCH SUMMARY \n"
            f"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ \n"
            f"- Epoch number: {epoch_number:04d} / {total_epochs:04d} \n"
            f"- Training loss:   {train_loss:04f} \n"
            f"- Validation loss: {validation_loss} \n"
            f"- Epoch took: {epoch_time} \n"
            f"- Time since start: {overall_time} \n"
            f"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ \n"
            f"Saving to {saved_model_path}. \n"
            f"Output @ {trainer.state.output_folder} \n"
        )

        logging.info(logging_str)


def add_progress_bar_handler(*engines):
    """
    Adds a progress bar to each engine. Keeps track of a running
    average of the loss as well.
    
    Usage::

    .. code-block:: python

        tr_engine, val_engine = ...
        add_progress_bar_handler(tr_engine, val_engine)
    """
    for engine in engines:
        output_transform = lambda x: x['loss']
        RunningAverage(output_transform=output_transform).attach(engine, 'avg_loss')
        ProgressBar().attach(engine, ['avg_loss'])
        

def add_tensorboard_handler(tensorboard_folder, engine, every_iteration=False):
    """
    Every key in engine.state.epoch_history[-1] is logged to TensorBoard.
    
    Args:
        tensorboard_folder (str): Where the tensorboard logs should go.
        trainer (ignite.Engine): The engine to log.
        every_iteration (bool, optional): Whether to also log the values at every 
          iteration.
    """
    writer = SummaryWriter(tensorboard_folder)

    @engine.on(ValidationEvents.VALIDATION_COMPLETED)
    def log_to_tensorboard(engine):
        for key in engine.state.epoch_history:
            writer.add_scalar(
                key, engine.state.epoch_history[key][-1], engine.state.epoch)

    if every_iteration:
        @engine.on(Events.ITERATION_COMPLETED)
        def log_iteration_to_tensorboard(engine):
            for key in engine.state.iter_history:
                writer.add_scalar(
                    key, engine.state.iter_history[key][-1], engine.state.iteration)



import torch
import yaml
import json

from ...ml import SeparationModel
from ...datasets import transforms as tfm
from ...core.migration import SafeModelLoader

OMITTED_TRANSFORMS = (
    tfm.GetExcerpt,
    tfm.MagnitudeWeights,
    tfm.SumSources,
    tfm.Cache,
    tfm.IndexSources,
)


class DeepMixin:
    def load_model(self, model_path, device='cpu'):
        """
        Loads the model at specified path `model_path`. Uses GPU if
        available.

        Args:
            model_path (str): path to model saved as SeparationModel.
            device (str or torch.Device): loads model on CPU or GPU. Defaults to
              'cuda'.

        Returns:
            model (SeparationModel): Loaded model, nn.Module
            metadata (dict): metadata associated with model, used for making
            the input data into the model.
        """
        safe_loader = SafeModelLoader()
        model_dict = safe_loader.load(model_path, 'cpu')
        metadata = model_dict['metadata']

        model = SeparationModel(metadata['config'])
        model.load_state_dict(model_dict['state_dict'])
        device = device if torch.cuda.is_available() else 'cpu'

        self.device = device

        model = model.to(device).eval()
        self.model = model
        self.config = metadata['config']
        self.metadata.update(metadata)
        self.transform = self._get_transforms(metadata['train_dataset']['transforms'])

    @staticmethod
    def _get_transforms(loaded_tfm):
        """
        Look through the loaded transforms and omits any that are in 
        `OMITTED_TRANSFORMS`.
        
        Args:
            loaded_tfm (Transform): A Transform from `nussl.datasets.transforms`.
        
        Returns:
            Transform: If the transform was a Compose, this returns a new Compose that
              omits the transforms listed in `OMITTED_TRANSFORMS`.
        """
        if isinstance(loaded_tfm, tfm.Compose):
            transform = []
            for _tfm in loaded_tfm.transforms:
                if not isinstance(_tfm, OMITTED_TRANSFORMS):
                    transform.append(_tfm)
            transform = tfm.Compose(transform)
        else:
            if not isinstance(loaded_tfm, OMITTED_TRANSFORMS):
                transform = loaded_tfm
            else:
                transform = None
        return transform

    def _get_input_data_for_model(self, extra_data=None):
        """
        Sets up the audio signal with the appropriate STFT parameters and runs it
        through the transform found in the metadata.

        Args:
            extra_data: A dictionary containing any additional data that will 
              be merged with the output dictionary.
        
        Returns:
            dict: Data dictionary to pass into the model.
        """
        extra_data = {} if extra_data is None else extra_data
        if self.metadata['sample_rate'] is not None:
            if self.audio_signal.sample_rate != self.metadata['sample_rate']:
                self.audio_signal.resample(self.metadata['sample_rate'])

        self.audio_signal.stft_params = self.metadata['stft_params']
        self.audio_signal.stft()

        data = {'mix': self.audio_signal}
        data.update(extra_data)
        data = self.transform(data)

        for key in data:
            if torch.is_tensor(data[key]):
                data[key] = data[key].unsqueeze(0).to(self.device).float()
                if self.metadata['num_channels'] == 1:
                    # then each channel is processed indep
                    data[key] = data[key].transpose(0, self.channel_dim)
        self.input_data = data
        return self.input_data

    def get_metadata(self, to_str=False, **kwargs):
        """
        Gets the metadata associated with this model.
        Args:
            to_str (bool): If True, will return a string, else will return dict.
            for_upload (bool): If True, will scrub metadata for uploading to EFZ.

        Returns:
            (str) or (dict) containing metadata.
        """
        for_upload = kwargs.get('for_upload', False)
        truncate_loss = kwargs.get('truncate_loss', False)
        metadata = getattr(self, 'metadata', None)

        if metadata is None:
            raise ValueError('Could not find associated metadata.')

        if for_upload:
            # remove paths
            keys = ['train_dataset', 'val_dataset']
            for k in keys:
                if k in metadata:
                    metadata[k].pop('folder')

        if for_upload or truncate_loss:
            if 'trainer.state.epoch_history' in metadata:
                loss_history = metadata.pop('trainer.state.epoch_history')
                metadata['final_loss'] = {k: float(v[-1]) for k, v in loss_history.items()}

        if isinstance(metadata['config'], str):
            metadata['config'] = json.loads(metadata['config'])
        metadata['separation_class'] = type(self).__name__
        metadata['model_name'] = metadata['config']['name']

        if to_str:
            return yaml.dump(metadata, indent=4)
        else:
            return metadata



"""
Provides utilities for running nussl algorithms that do not belong to
any specific algorithm or that are shared between algorithms.
"""

import warnings

import numpy as np
import torch
import random
from .. import musdb
import librosa
from . import constants

import os
from contextlib import contextmanager

def seed(random_seed, set_cudnn=False):
    """
    Seeds all random states in nussl with the same random seed
    for reproducibility. Seeds ``numpy``, ``random`` and ``torch``
    random generators. 

    For full reproducibility, two further options must be set
    according to the torch documentation:

    https://pytorch.org/docs/stable/notes/randomness.html

    To do this, ``set_cudnn`` must be True. It defaults to 
    False, since setting it to True results in a performance
    hit.

    Args:
        random_seed (int): integer corresponding to random seed to 
        use.
        set_cudnn (bool): Whether or not to set cudnn into determinstic
        mode and off of benchmark mode. Defaults to False.
    """

    torch.manual_seed(random_seed)
    np.random.seed(random_seed)
    random.seed(random_seed)

    if set_cudnn:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def find_peak_indices(input_array, n_peaks, min_dist=None, do_min=False, threshold=0.5):
    """
    This function will find the indices of the peaks of an input n-dimensional numpy array.
    This can be configured to find max or min peak indices, distance between the peaks, and
    a lower bound, at which the algorithm will stop searching for peaks (or upper bound if
    searching for max). Used exactly the same as :func:`find_peak_values`.

    This function currently only accepts 1-D and 2-D numpy arrays.

    Notes:
        * This function only returns the indices of peaks. If you want to find peak values,
        use :func:`find_peak_values`.

        * min_dist can be an int or a tuple of length 2.
            If input_array is 1-D, min_dist must be an integer.
            If input_array is 2-D, min_dist can be an integer, in which case the minimum
            distance in both dimensions will be equal. min_dist can also be a tuple if
            you want each dimension to have a different minimum distance between peaks.
            In that case, the 0th value in the tuple represents the first dimension, and
            the 1st value represents the second dimension in the numpy array.

    Args:
        input_array: a 1- or 2- dimensional numpy array that will be inspected.
        n_peaks: (int) maximum number of peaks to find
        min_dist: (int) minimum distance between peaks. Default value: len(input_array) / 4
        do_min: (bool) if True, finds indices at minimum value instead of maximum
        threshold: (float) the value (scaled between 0.0 and 1.0)

    Returns:
        peak_indices: (list) list of the indices of the peak values

    """
    input_array = np.array(input_array, dtype=float)

    if input_array.ndim > 2:
        raise ValueError('Cannot find peak indices on data greater than 2 dimensions!')

    is_1d = input_array.ndim == 1
    zero_dist = zero_dist0 = zero_dist1 = None
    min_dist = len(input_array) // 4 if min_dist is None else min_dist

    if is_1d:
        zero_dist = min_dist
    else:
        if type(min_dist) is int:
            zero_dist0 = zero_dist1 = min_dist
        elif len(min_dist) == 1:
            zero_dist0 = zero_dist1 = min_dist[0]
        else:
            zero_dist0, zero_dist1 = min_dist

    # scale input_array between [0.0, 1.0]
    input_array -= np.min(input_array)
    input_array /= np.max(input_array)

    # flip sign if doing min
    input_array = -input_array if do_min else input_array

    # throw out everything below threshold
    input_array = np.multiply(input_array, (input_array >= threshold))

    # check to make sure we didn't throw everything out
    if np.size(np.nonzero(input_array)) == 0:
        raise ValueError('Threshold set incorrectly. No peaks above threshold.')
    if np.size(np.nonzero(input_array)) < n_peaks:
        warnings.warn('Threshold set such that there will be less peaks than n_peaks.')

    peak_indices = []
    for i in range(n_peaks):
        # np.unravel_index for 2D indices e.g., index 5 in a 3x3 array should be (1, 2)
        # Also, wrap in list for duck typing
        cur_peak_idx = list(np.unravel_index(
            np.argmax(input_array.flatten()), input_array.shape
        ))

        # zero out peak and its surroundings
        if is_1d:
            cur_peak_idx = cur_peak_idx[0]
            peak_indices.append(cur_peak_idx)
            lower, upper = _set_array_zero_indices(cur_peak_idx, zero_dist, len(input_array))
            input_array[lower:upper] = 0
        else:
            peak_indices.append(cur_peak_idx)
            lower0, upper0 = _set_array_zero_indices(cur_peak_idx[0], zero_dist0,
                                                     input_array.shape[0])
            lower1, upper1 = _set_array_zero_indices(cur_peak_idx[1], zero_dist1,
                                                     input_array.shape[1])
            input_array[lower0:upper0, lower1:upper1] = 0

        if np.sum(input_array) == 0.0:
            break

    return peak_indices


def _set_array_zero_indices(index, zero_distance, max_len):
    lower = index - zero_distance
    upper = index + zero_distance + 1
    lower = 0 if lower < 0 else lower
    upper = max_len if upper >= max_len else upper
    return int(lower), int(upper)


def complex_randn(shape):
    """
    Returns a complex-valued numpy array of random values with shape :param:`shape`.

    Args:
        shape (tuple): Tuple of ints that will be the shape of the resultant complex numpy array.

    Returns:
        (:obj:`np.ndarray`): a complex-valued numpy array of random values with shape `shape`
    """
    return np.random.randn(*shape) + 1j * np.random.randn(*shape)


def _get_axis(array, axis_num, i):
    """
    Will get index 'i' along axis 'axis_num' using np.take.

    Args:
        array (:obj:`np.ndarray`): Array to fetch axis of.
        axis_num (int): Axes to retrieve.
        i (int): Index to retrieve.

    Returns:
        The value at index :param:`i` along axis :param:`axis_num`
    """

    return np.take(array, i, axis_num)


def _slice_along_dim(data, dim, start, end):
    """
    Takes a slice of data along a dim between a start and an end. Agnostic to
    whether the data is a numpy array or a torch tensor.
    
    Args:
        data (np.ndarray or torch.Tensor): Data to slice.
        dim (int): Dimension along which to do the slicing.
        start (int): Start of the slice.
        end (int): End of the slice
    """
    if dim > 3:
        raise ValueError("Unsupported for dim > 4")
    if dim >= len(data.shape):
        raise ValueError(f"dim {dim} too high for data.shape {data.shape}!")

    if dim == 0:
        return data[start:end, ...]
    elif dim == 1:
        return data[:, start:end, ...]
    elif dim == 2:
        return data[:, :, start:end, ...]
    elif dim == 3:
        return data[:, :, :, start:end, ...]


def _format(string):
    """ Formats a class name correctly for checking function and class names.
        Strips all non-alphanumeric chars and makes lowercase.
    """
    return ''.join(list(filter(str.isalnum, string))).lower()


def musdb_track_to_audio_signals(track):
    """
    Converts a musdb track to a dictionary of AudioSignal objects.
    
    Args:
        track (musdb.audio_classes.MultiTrack): MultiTrasack object 
            containing stems that will each be turned into AudioSignal
            objects.

    Returns:
        (2-tuple): tuple containing the mixture AudioSignal and a dictionary of
            the sources.
    """
    # lazy load to prevent circular imports
    from .audio_signal import AudioSignal

    mixture = AudioSignal(audio_data_array=track.audio, sample_rate=track.rate)
    mixture.path_to_input_file = track.name
    stems = track.stems
    sources = {}

    for k, v in sorted(track.sources.items(), key=lambda x: x[1].stem_id):
        sources[k] = AudioSignal(
            audio_data_array=stems[v.stem_id],
            sample_rate=track.rate
        )
        sources[k].path_to_input_file = f'musdb/{track.name}_{k}.wav'

    return mixture, sources


def audio_signals_to_musdb_track(mixture, sources_dict, targets_dict):
    """
    Converts :class:`AudioSignal` objects to ``musdb`` :class:`Track` objects that
    contain the mixture, the ground truth sources, and the targets for use with the ``mus_eval``
    implementation of BSS-Eval and ``musdb``.

    See Also:
        * More information on ``musdb``:  `Github<https://github.com/sigsep/sigsep-mus-db>`
            and `documentation<http://musdb.readthedocs.io/>`
        * More information on ``mus_eval``: `Github<https://github.com/sigsep/sigsep-mus-eval>`
            and `documentation<https://sigsep.github.io/sigsep-mus-eval/>`
        * :class:`BSSEvalV4` for *nussl*'s interface to BSS-Eval v4.

    Examples:
        .. code-block:: python
            :linenos:
            import nussl
            signal = nussl.AudioSignal(nussl.efz_utils.download_audio_file('HistoryRepeating.wav'))

            repet = nussl.Repet(signal)
            repet.run()

            bg, fg = repet.make_audio_signals()

            src_dict = {'vocals': fg, 'accompaniment': bg}
            target = nussl.core.constants.STEM_TARGET_DICT
            track = nussl.utils.audio_signals_to_musdb_track(signal, src_dict, target)

    Args:
        mixture (:class:`AudioSignal`): The :class:`AudioSignal` object that contains the mixture.
        sources_dict (dict): Dictionary where the keys are the labels for the sources and values
            are the associated :class:`AudioSignal` objects.
        targets_dict (dict): Dictionary where the keys are the labels for the sources (as above)
            and the values are weights.

    Returns:
        (:obj:`musdb.MultiTrack`) populated as specified by inputs.
    """
    verify_audio_signal_list_strict(list(sources_dict.values()) + [mixture])

    path = mixture.path_to_input_file if mixture.path_to_input_file else "None"
    fname = mixture.file_name if mixture.file_name else "None"
    track = musdb.audio_classes.MultiTrack(path=path, name=fname, is_wav=True)
    track.audio = mixture.audio_data.T
    track.rate = mixture.sample_rate

    stems = [track.audio]

    for name, target_srcs in list(targets_dict.items()):
        if name in sources_dict:
            stems.append(sources_dict[name].audio_data.T)

    track._stems = np.array(stems)
    return track


def verify_audio_signal_list_lax(audio_signal_list):
    """
    Verifies that an input (:param:`audio_signal_list`) is a list of :ref:`AudioSignal` objects.
    If not so, attempts to correct the list (if possible) and returns the corrected list.

    Args:
        audio_signal_list (list): List of :ref:`AudioSignal` objects

    Returns:
        audio_signal_list (list): Verified list of :ref:`AudioSignal` objects.

    """
    # Lazy load to prevent a circular reference upon initialization
    from .audio_signal import AudioSignal

    if isinstance(audio_signal_list, AudioSignal):
        audio_signal_list = [audio_signal_list]
    elif isinstance(audio_signal_list, list):
        if not all(isinstance(s, AudioSignal) for s in audio_signal_list):
            raise ValueError('All input objects must be AudioSignal objects!')
        if not all(s.has_data for s in audio_signal_list):
            raise ValueError('All AudioSignal objects in input list must have data!')
    else:
        raise ValueError(
            'audio_signal_list must be a list of or a single AudioSignal objects!')

    return audio_signal_list


def verify_audio_signal_list_strict(audio_signal_list):
    """
    Verifies that an input (:param:`audio_signal_list`) is a list of :ref:`AudioSignal` objects and
    that they all have the same sample rate and same number of channels. If not true,
    attempts to correct the list (if possible) and returns the corrected list.

    Args:
        audio_signal_list (list): List of :ref:`AudioSignal` objects

    Returns:
        audio_signal_list (list): Verified list of :ref:`AudioSignal` objects, that all have
        the same sample rate and number of channels.

    """
    audio_signal_list = verify_audio_signal_list_lax(audio_signal_list)

    if not all(audio_signal_list[0].sample_rate == s.sample_rate for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same sample rate!')

    if not all(audio_signal_list[0].num_channels == s.num_channels for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same number of channels!')

    if not all(audio_signal_list[0].signal_length == s.signal_length for s in audio_signal_list):
        raise ValueError('All input AudioSignal objects must have the same signal length!')

    return audio_signal_list


def visualize_gradient_flow(named_parameters, n_bins=50):
    """
    Visualize the gradient flow through the named parameters of a PyTorch model. 

    Plots the gradients flowing through different layers in the net during training.
    Can be used for checking for possible gradient vanishing / exploding problems.

    Usage: Plug this function in Trainer class after loss.backwards() as 
    "visualize_gradient_flow(self.model.named_parameters())" to visualize 
    the gradient flow
    
    Args:
        named_parameters (generator): Generator object yielding name and parameters
          for each layer in a PyTorch model.
        n_bins (int): Number of bins to use for each histogram. Defaults to 50.
    """
    import matplotlib.pyplot as plt

    data = []

    for n, p in named_parameters:
        if p.requires_grad and "bias" not in n:
            if p.grad is not None:
                _data = p.grad.cpu().data.numpy().flatten()
                lower = np.percentile(_data, 10)
                upper = np.percentile(_data, 90)
                _data = _data[_data >= lower]
                _data = _data[_data <= upper]
                n = n.split('layers.')[-1]
                data.append((n, _data, np.abs(_data).mean()))

    _data = [d[1] for d in sorted(data, key=lambda x: x[-1])]
    _names = [d[0] for d in sorted(data, key=lambda x: x[-1])]

    plt.hist(_data, len(_data) * n_bins, histtype='step', fill=False,
             stacked=True, label=_names)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2)

def visualize_spectrogram(audio_signal, ch=0, do_mono=False, x_axis='time', 
                          y_axis='linear',  **kwargs):
    """
    Wrapper around `librosa.display.specshow` for usage with AudioSignals.
    
    Args:
        audio_signal (AudioSignal): AudioSignal to plot
        ch (int, optional): Which channel to plot. Defaults to 0.
        do_mono (bool, optional): Make the AudioSignal mono. Defaults to False.
        x_axis (str, optional): x_axis argument to librosa.display.specshow. Defaults to 'time'.
        y_axis (str, optional): y_axis argument to librosa.display.specshow. Defaults to 'linear'.
        kwargs: Additional keyword arguments to librosa.display.specshow.
    """
    import librosa.display

    if do_mono:
        audio_signal = audio_signal.to_mono(overwrite=False)
    
    data = librosa.amplitude_to_db(np.abs(audio_signal.stft()), ref=np.max)
    librosa.display.specshow(data[..., ch], x_axis=x_axis, y_axis=y_axis, 
        sr=audio_signal.sample_rate, hop_length=audio_signal.stft_params.hop_length,
        **kwargs)

def visualize_waveform(audio_signal, ch=0, do_mono=False, x_axis='time', **kwargs):
    """
    Wrapper around `librosa.display.waveplot` for usage with AudioSignals.
    
    Args:
        audio_signal (AudioSignal): AudioSignal to plot
        ch (int, optional): Which channel to plot. Defaults to 0.
        do_mono (bool, optional): Make the AudioSignal mono. Defaults to False.
        x_axis (str, optional): x_axis argument to librosa.display.waveplot. Defaults to 'time'.
        kwargs: Additional keyword arguments to librosa.display.waveplot.
    """
    import librosa.display
    import matplotlib.pyplot as plt

    if do_mono:
        audio_signal = audio_signal.to_mono(overwrite=False)
    
    data = np.asfortranarray(audio_signal.audio_data[ch])
    librosa.display.waveplot(data, sr=audio_signal.sample_rate, x_axis=x_axis, **kwargs)
    plt.ylabel('Amplitude')

def visualize_sources_as_waveform(audio_signals, ch=0, do_mono=False, x_axis='time', 
                                  colors=None, alphas=None, show_legend=True, **kwargs):
    """
    Visualizes a dictionary or list of sources with overlapping waveforms with transparency.
    
    The labels of each source are either the key, if a dictionary, or the 
    path to the input audio file, if a list.
    
    Args:
        audio_signals (list or dict): List or dictionary of audio signal objects to be
          plotted.
        ch (int, optional): Which channel to plot. Defaults to 0.
        do_mono (bool, optional): Make each AudioSignal mono. Defaults to False.
        x_axis (str, optional): x_axis argument to librosa.display.waveplot. Defaults to 'time'.
        colors (list, optional): Sequence of colors to use for each signal. 
          Defaults to None, which uses the default matplotlib color cycle.
        alphas (list, optional): Sequence of alpha transparency to use for each signal. 
          Defaults to None.
        kwargs: Additional keyword arguments to librosa.display.waveplot.
    """
    import matplotlib.pyplot as plt

    if isinstance(audio_signals, list):
        audio_signals = {
            f'{i}:{a.path_to_input_file}': a 
            for i, a in enumerate(audio_signals)
        }

    sorted_keys = sorted(
        audio_signals.keys(),
        key=lambda k: audio_signals[k].rms().mean(),
        reverse=True
    )

    alphas = (
        np.linspace(0.25, .75, len(audio_signals)) 
        if alphas is None else alphas
    )
    colors = (
        plt.rcParams['axes.prop_cycle'].by_key()['color'] 
        if colors is None else colors
    )

    for i, key in enumerate(sorted_keys):
        val = audio_signals[key]
        color = colors[i % len(audio_signals)]
        visualize_waveform(val, ch=ch, do_mono=do_mono, x_axis=x_axis, 
                           alpha=alphas[i % len(audio_signals)],
                           label=key, color=color)

    if show_legend:
        plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2)

def visualize_sources_as_masks(audio_signals, ch=0, do_mono=False, x_axis='time', 
                               y_axis='linear', db_cutoff=-60, colors=None, alphas=None, 
                               alpha_amount=1.0, show_legend=True, **kwargs):
    """
    Visualizes a dictionary or list of sources with overlapping waveforms with transparency.
    
    The labels of each source are either the key, if a dictionary, or the 
    path to the input audio file, if a list.
    
    Args:
        audio_signals (list or dict): List or dictionary of audio signal objects to be
          plotted.
        ch (int, optional): Which channel to plot. Defaults to 0.
        do_mono (bool, optional): Make each AudioSignal mono. Defaults to False.
        x_axis (str, optional): x_axis argument to librosa.display.waveplot. Defaults to 'time'.
        colors (list, optional): Sequence of colors to use for each signal. 
          Defaults to None, which uses the default matplotlib color cycle.
        alphas (list, optional): Sequence of alpha transparency to use for each signal. 
          Defaults to None.
        kwargs: Additional keyword arguments to librosa.display.specshow.
    """
    import matplotlib.pyplot as plt
    import matplotlib
    import librosa.display
    from .. import datasets

    if isinstance(audio_signals, list):
        audio_signals = {
            f'{i}:{a.path_to_input_file}': a 
            for i, a in enumerate(audio_signals)
        }

    if do_mono:
        for key in audio_signals:
            audio_signals[key] = audio_signals[key].to_mono()

    sorted_keys = sorted(
        audio_signals.keys(),
        key=lambda k: audio_signals[k].rms().mean(),
        reverse=True
    )

    source_names = sorted(list(audio_signals.keys()))
    mix = sum(audio_signals.values())
    data = {
        'mix': mix,
        'sources': audio_signals
    }
    data = datasets.transforms.PhaseSensitiveSpectrumApproximation()(data)

    colors = (
        plt.rcParams['axes.prop_cycle'].by_key()['color'] 
        if colors is None else colors
    )

    # construct each image with alpha values
    masks = data['source_magnitudes'] / (np.maximum(
            data['mix_magnitude'][..., None], data['source_magnitudes']) 
            + constants.EPSILON
        )
    legend_elements = []

    silence_mask = librosa.amplitude_to_db(np.abs(mix.stft()), ref=np.max) > db_cutoff
    masks *= silence_mask[..., None]

    y_coords = librosa.display.__mesh_coords(y_axis, None, masks.shape[0], 
        sr=mix.sample_rate, hop_length=mix.stft_params.hop_length)
    x_coords = librosa.display.__mesh_coords(x_axis, None, masks.shape[1],
        sr=mix.sample_rate, hop_length=mix.stft_params.hop_length)

    extent = [x_coords.min(), x_coords.max(), y_coords.min(), y_coords.max()]

    for j, key in enumerate(sorted_keys):
        i = source_names.index(key)
        mask = masks[..., ch, i]
        color = colors[j % len(colors)]

        cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            'custom', ['white', color])
        image = cmap(mask)
        image[:, :, -1] = mask ** alpha_amount
        plt.imshow(image, origin='lower', aspect='auto', 
            interpolation='none', extent=extent)

        legend_elements.append(
            matplotlib.patches.Patch(facecolor=color, label=key))

    axes = librosa.display.__check_axes(None)

    axes.set_xlim(x_coords.min(), x_coords.max())
    axes.set_ylim(y_coords.min(), y_coords.max())

    # Set up axis scaling
    librosa.display.__scale_axes(axes, x_axis, 'x')
    librosa.display.__scale_axes(axes, y_axis, 'y')

    # Construct tickers and locators
    librosa.display.__decorate_axis(axes.xaxis, x_axis)
    librosa.display.__decorate_axis(axes.yaxis, y_axis)

    
    if show_legend:
        plt.legend(handles=legend_elements,  
            bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=2)

@contextmanager
def _close_temp_files(tmpfiles):
    """
    Utility function for creating a context and closing all temporary files
    once the context is exited. For correct functionality, all temporary file
    handles created inside the context must be appended to the ```tmpfiles```
    list.

    This function is taken wholesale from Scaper.

    Args:
        tmpfiles (list): List of temporary file handles
    """
    def _close():
        for t in tmpfiles:
            try:
                t.close()
                os.unlink(t.name)
            except:
                pass
    try:
        yield
    except:
        _close()
        raise
    _close()



import numpy as np
from scipy.ndimage.filters import convolve
from scipy.ndimage import maximum_filter, gaussian_filter

from .. import MaskSeparationBase, SeparationException
from ..benchmark import HighLowPassFilter
from ... import AudioSignal
from ... import vamp_imported
import numpy as np
import scipy.signal

if vamp_imported:
    import vamp

# function for generating the vocal chord impulse response
def rosenmodel(t1, t2, fs):
    """
    This model for generating singing vowel sounds from sine tones comes
    from:

    https://simonl02.users.greyc.fr/files/Documents/Teaching/SignalProcessingLabs/lab3.pdf

    The above will be referred to throughout these docstrings as THE DOCUMENT.

    Original author: Fatemeh Pishdadian

    The arguments to the fucntions throughout this text follow the 
    signatures laid out in THE DOCUMENT.

    This is used in Melodia to generate the melody signal to produce a 
    mask.

    Equation 2 in THE DOCUMENT.
    """

    N1 = np.floor(t1 * fs)
    N2 = np.floor(t2 * fs)

    samp_vec1 = np.arange(N1+1)
    samp_vec2 = np.arange(N1,N1+N2+1)

    ir_func1 =  0.5 * (1 - np.cos((np.pi * samp_vec1)/N1))
    ir_func2 = np.cos(np.pi * (samp_vec2 - N1)/(2 * N2))

    vchord_filt = np.concatenate((ir_func1,ir_func2))

    return vchord_filt


# function for computing the denominator coeffs of the vocal cavity filter transfer function
def oral_cavity_filt(pole_amps, pole_freqs,fs):
    """
    This model for generating singing vowel sounds from sine tones comes
    from:

    https://simonl02.users.greyc.fr/files/Documents/Teaching/SignalProcessingLabs/lab3.pdf

    The above will be referred to throughout these docstrings as THE DOCUMENT.

    Original author: Fatemeh Pishdadian

    The arguments to the fucntions throughout this text follow the 
    signatures laid out in THE DOCUMENT.

    This is used in Melodia to generate the melody signal to produce a 
    mask.

    Solves "Q. Write a function to synthesize filter H(z)" in 
    THE DOCUMENT
    """
    num_pole_pair = len(pole_amps)
    poles = pole_amps * np.exp(1j * 2 * np.pi * pole_freqs / fs)
    poles_conj = np.conj(poles)

    denom_coeffs = 1
    for i in range(num_pole_pair):
        pole_temp = poles[i]
        pole_conj_temp = poles_conj[i]
        pole_pair_coeffs = np.convolve(np.array([1,-pole_temp]),np.array([1,-pole_conj_temp]))

        denom_coeffs = np.convolve(denom_coeffs, pole_pair_coeffs)
    return denom_coeffs

def _apply_vowel_filter(impulse_train, fs, t1=0.0075, t2=.013,
                     pole_amps=None, pole_freqs=None):
    """
    This model for generating singing vowel sounds from sine tones comes
    from:

    https://simonl02.users.greyc.fr/files/Documents/Teaching/SignalProcessingLabs/lab3.pdf

    The above will be referred to throughout these docstrings as THE DOCUMENT.

    Original author: Fatemeh Pishdadian

    The arguments to the fucntions throughout this text follow the 
    signatures laid out in THE DOCUMENT.

    This is used in Melodia to generate the melody signal to produce a 
    mask.
    
    Args:
        impulse_train (np.ndarray): Numpy array with data to be filtered
        fs (int): Sample rate of audio.
        t1 (float, optional): N1 in Equation 2 in THE DOCUMENT. Defaults to 0.0075.
        t2 (float, optional): N2 in Equation 2 in THE DOCUMENT. Defaults to .013.
        pole_amps (np.ndarray, optional): Pole amplitudes, see Figures 2-4 in THE DOCUMENT. 
          Defaults to None, which maps to E vowel.
        pole_freqs (np.ndarray, optional): Pole frequencies, see Figures 2-4 in THE DOCUMENT. 
          Defaults to None, which maps to E vowel
    
    Returns:
        np.ndarray: Filtered impulse train that should sound sort of like the desired 
          vowel.
    """
    if pole_amps is None:
        pole_amps = np.array([0.99,0.98,0.9,0.9])
    if pole_freqs is None:
        pole_freqs = np.array([800,1200,2800,3600])
        
    vchord_filt = rosenmodel(t1, t2, fs)
    vchord_out = np.convolve(impulse_train, vchord_filt)
        
    denom_coeffs = oral_cavity_filt(pole_amps, pole_freqs, fs)
    oral_out = scipy.signal.lfilter(
        np.array([1]), denom_coeffs, vchord_out)
    lip_out = np.real(scipy.signal.lfilter(
        np.array([1,-1]), np.array([1]), oral_out))
    
    lip_out = lip_out[:impulse_train.shape[0]]
    
    return np.real(lip_out)

class Melodia(MaskSeparationBase):
    """
    Implements melody extraction using Melodia [1].

    This needs Melodia installed as a vamp plugin, as well as having vampy for 
    Python installed. Install Melodia via: https://www.upf.edu/web/mtg/melodia.
    Note that Melodia can be used only for NON-COMMERCIAL use.

    References:

    [1] J. Salamon and E. Gómez, "Melody Extraction from Polyphonic Music Signals using 
        Pitch Contour Characteristics", IEEE Transactions on Audio, Speech and 
        Language Processing, 20(6):1759-1770, Aug. 2012.

    Args:
        input_audio_signal (AudioSignal object): The AudioSignal object that has the
          audio data that Melodia will be run on.

        high_pass_cutoff (optional, float): value (in Hz) for the high pass cutoff 
          filter.
        
        minimum_frequency (optional, float): minimum frequency in Hertz (default 55.0)

        maximum_frequency (optional, float): maximum frequency in Hertz (default 1760.0)

        voicing_tolerance (optional, float): Greater values will result in more pitch contours 
          included in the final melody. Smaller values will result in less pitch 
          contours included in the final melody (default 0.2).

        minimum_peak_salience (optional, float): a hack to avoid silence turning into junk 
          contours when analyzing monophonic recordings (e.g. solo voice with 
          no accompaniment). Generally you want to leave this untouched (default 0.0).

        num_overtones (optional, int): Number of overtones to use when creating 
          melody mask.

        apply_vowel_filter (optional, bool): Whether or not to apply a vowel filter
          on the resynthesized melody signal when masking.

        smooth_length (optional, int): Number of frames to smooth discontinuities in the
          mask.

        add_lower_octave (optional, fool): Use octave below fundamental frequency as well
          to take care of octave errors in pitch tracking, since we only care about
          the mask. Defaults to False.
        
        mask_type (optional, str): Type of mask to use.

        mask_threshold (optional, float): Threshold for mask to convert to binary.
    """

    def __init__(self, input_audio_signal, high_pass_cutoff=100, minimum_frequency=55.0,
                 maximum_frequency=1760.0, voicing_tolerance=0.2, minimum_peak_salience=0.0,
                 compression=0.5, num_overtones=40, apply_vowel_filter=False, smooth_length=5, 
                 add_lower_octave=False, mask_type='soft', mask_threshold=0.5):
        # lazy load vamp to check if it exists
        from ... import vamp_imported

        melodia_installed = False
        if vamp_imported:
            melodia_installed = 'mtg-melodia:melodia' in vamp.list_plugins()

        if not vamp_imported or not melodia_installed:
            self._raise_vamp_melodia_error()

        super().__init__(
            input_audio_signal=input_audio_signal,
            mask_type=mask_type,
            mask_threshold=mask_threshold
        )

        self.high_pass_cutoff = high_pass_cutoff
        self.minimum_frequency = float(minimum_frequency)
        self.maximum_frequency = float(maximum_frequency)
        self.voicing_tolerance = float(voicing_tolerance)
        self.minimum_peak_salience = float(minimum_peak_salience)
        self.compression = compression
        self.apply_vowel_filter = apply_vowel_filter
        self.add_lower_octave = add_lower_octave

        self.melody = None
        self.melody_signal = None
        self.timestamps = None

        self.num_overtones = num_overtones
        self.smooth_length = smooth_length

    def _raise_vamp_melodia_error(self):
        raise SeparationException(
            '\n**~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~**'
            '\n*          Are Vamp and Melodia installed correctly?          *'
            '\n* Check https://bit.ly/2DXbrAk for installation instructions! *'
            '\n**~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~*~**')

    def extract_melody(self):
        """
        Extracts melody from the audio using the melodia vamp plugin. Uses arguments kept 
        in self:
        
        - `self.minimum_frequency` (default: 55 Hz)
        - `self.maximum_frequency` (default: 1760 Hz)
        - `self.voicing_tolerance` (default: 0.2)
        - `self.minimum_peak_salience` (default: 0.0)

        This function sets two class members used in other parts:

        - `self.melody`: (numpy array) contains the melody in Hz for every timestep 
          (0 indicates no voice).
        - `self.timestamps`: (numpy array) contains the timestamps for each melody note
        """

        params = {
            'minfqr': self.minimum_frequency,
            'maxfqr': self.maximum_frequency,
            'voicing': self.voicing_tolerance,
            'minpeaksalience': self.minimum_peak_salience
        }

        data = vamp.collect(self.audio_signal.audio_data, self.sample_rate,
                            "mtg-melodia:melodia", parameters=params)

        _, melody = data['vector']
        hop = 128. / 44100.  # hard coded hop in Melodia vamp plugin, converting it to frames.
        timestamps = 8 * hop + np.arange(len(melody)) * hop
        melody[melody < 0] = 0
        self.melody = melody
        self.timestamps = timestamps

    def create_melody_signal(self, num_overtones):
        """
        Adapted from Melosynth by Justin Salamon: https://github.com/justinsalamon/melosynth. 
        To mask the mixture, we need to identify time-frequency bins that belong to the 
        melody. Melodia outputs only the fundamental frequency of the melodic line. 
        To construct the mask we take the fundamental frequency and add all the 
        overtones of it (up to num_overtones) to the mask. The melody is faded in and 
        out at onsets and offsets to make the separation sound more natural 
        (hard-coded by transition_length).

        Args:

            num_overtones (int): Number of overtones to expand out to build the mask.

        """

        if self.timestamps[0] > 0:
            estimated_hop = np.median(np.diff(self.timestamps))
            previous_time = max(self.timestamps[0] - estimated_hop, 0)
            self.timestamps = np.insert(self.timestamps, 0, previous_time)
            self.melody = np.insert(self.melody, 0, 0)

        sample_rate = self.audio_signal.sample_rate
        melody_signal = []
        transition_length = .001  # duration for fade in/out and frequency interpretation
        phase = np.zeros(num_overtones)
        previous_frequency = 0
        previous_time = 0

        overtone_weights = np.ones(num_overtones)

        for time, frequency in zip(self.timestamps, self.melody):
            if self.add_lower_octave:
                frequency = frequency / 2 
            # taking care of octave errors since we only care about masking
            num_samples = int(np.round((time - previous_time) * sample_rate))
            if num_samples > 0:
                num_transition_samples = float(
                    min(np.round(transition_length * sample_rate), num_samples))
                frequency_series = np.ones(num_samples) * previous_frequency

                if previous_frequency > 0 and frequency > 0:
                    frequency_series += np.minimum(
                        np.arange(num_samples) / num_transition_samples, 1) * \
                                        (frequency - previous_frequency)
                elif frequency > 0:
                    frequency_series = np.ones(num_samples) * frequency

                samples = np.zeros(num_samples)

                for overtone in range(num_overtones):
                    overtone_num = overtone + 1
                    phasors = 2 * np.pi * overtone_num * frequency_series / float(sample_rate)
                    phases = phase[overtone] + np.cumsum(phasors)
                    samples += overtone_weights[overtone] * np.sign(np.sin(phases))
                    phase[overtone] = phases[-1]

                if previous_frequency == 0 and frequency > 0:
                    samples *= np.minimum(np.arange(num_samples) / num_transition_samples, 1)
                elif previous_frequency > 0 and frequency == 0:
                    samples *= np.maximum(1 - np.arange(num_samples) / num_transition_samples, 0)
                elif previous_frequency == 0 and frequency == 0:
                    samples *= 0

                melody_signal.extend(samples)

            previous_frequency = frequency
            previous_time = time

        melody_signal = np.asarray(melody_signal)
        if self.apply_vowel_filter:
            melody_signal = _apply_vowel_filter(melody_signal, sample_rate)

        melody_signal /= float(max(np.max(melody_signal), 1e-7))
        melody_signal = [melody_signal for _ in range(self.audio_signal.num_channels)]
        melody_signal = np.asarray(melody_signal)
        melody_signal = melody_signal[:, 0:self.audio_signal.signal_length]
        melody_signal = AudioSignal(
            audio_data_array=melody_signal,
            sample_rate=sample_rate,
            stft_params=self.audio_signal.stft_params
        )

        self.melody_signal = melody_signal
        return melody_signal

    def create_harmonic_mask(self, melody_signal):
        """
        Creates a harmonic mask from the melody signal. The mask is smoothed to reduce 
        the effects of discontinuities in the melody synthesizer.
        """
        stft = np.abs(melody_signal.stft())

        # Need to threshold the melody stft since the synthesized
        # F0 sequence overtones are at different weights.
        stft = stft ** self.compression
        stft /= np.maximum(np.max(stft, axis=1, keepdims=True), 1e-7)

        mask = np.empty(self.stft.shape)

        # Smoothing the mask row-wise using a low-pass filter to
        # get rid of discontuinities in the mask.
        kernel = np.full((1, self.smooth_length), 1 / self.smooth_length)
        for ch in range(self.audio_signal.num_channels):
            mask[..., ch] = convolve(stft[..., ch], kernel)
        return mask

    def run(self):
        high_low = HighLowPassFilter(self.audio_signal, self.high_pass_cutoff)
        high_pass_masks = high_low.run()

        # separate the mixture foreground melody by masking
        if self.melody_signal is None:
            self.extract_melody()
            self.create_melody_signal(self.num_overtones)

        foreground_mask = self.create_harmonic_mask(self.melody_signal)
        foreground_mask = self.MASKS['soft'](foreground_mask)

        foreground_mask = foreground_mask
        background_mask = foreground_mask.invert_mask()

        _masks = np.stack(
            [background_mask.mask, foreground_mask.mask], axis=-1)

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)

            if i == 0:
                mask_data = np.maximum(mask_data, high_pass_masks[i].mask)
            elif i == 1:
                mask_data = np.minimum(mask_data, high_pass_masks[i].mask)

            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks



import torch

from ..base import MaskSeparationBase, DeepMixin, SeparationException
from ... import ml


class DeepMaskEstimation(DeepMixin, MaskSeparationBase):
    """
    Separates an audio signal using the masks produced by a deep model for every 
    time-frequency point. It expects that the model outputs a dictionary where one
    of the keys is 'masks'. This uses the `DeepMixin` class to load the model
    and set the audio signal's parameters to be appropriate for the model.
    
    Args:
        input_audio_signal: (AudioSignal`) An AudioSignal object containing the 
          mixture to be separated.        
        model_path (str, optional): Path to the model that will be used. Can be None, 
          so that you can initialize a class and load the model later.  
          Defaults to None.
        extra_data: A dictionary containing any additional data that will 
          be merged with the output dictionary. This can come from a dataset,
          or contain a query, etc.
        device (str, optional): Device to put the model on. Defaults to 'cpu'.
        **kwargs (dict): Keyword arguments for MaskSeparationBase.
    """
    def __init__(self, input_audio_signal, model_path=None, device='cpu', 
                 extra_data=None, **kwargs):
        super().__init__(input_audio_signal, **kwargs)
        if model_path is not None:
            self.load_model(model_path, device=device)
        self.model_output = None
        self.extra_data = extra_data
        # audio channel dimension in a mask estimation model
        self.channel_dim = -1

    def forward(self):
        input_data = self._get_input_data_for_model(self.extra_data)
        with torch.no_grad():
            output = self.model(input_data)
            if 'mask' not in output:
                raise SeparationException(
                    "This model is not a deep mask estimation model! "
                    "Did not find 'mask' key in output dictionary.")
            masks = output['mask']
            # swap back batch and sample dims
            if self.metadata['num_channels'] == 1:
                masks = masks.transpose(0, -2)
            masks = masks.squeeze(0).transpose(0, 1)
            masks = masks.cpu().data.numpy()
        self.model_output = output
        return masks

    def run(self, masks=None):
        self.result_masks = []

        if masks is None:
            masks = self.forward()

        for i in range(masks.shape[-1]):
            mask_data = masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = masks[..., i] == masks.max(axis=-1)
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)
        
        return self.result_masks

    def confidence(self, approach='silhouette_confidence', num_sources=2, **kwargs):
        """
        In embedding-based separation algorithms, we can compute a confidence
        measure based on the clusterability of the embedding space. This can
        be used if the model also computes an embedding alongside the estimates
        (e.g. as in Chimera models.) 
        
        Args:
            approach (str, optional): What approach to use for getting the confidence 
              measure. Options are 'jensen_shannon_confidence', 'posterior_confidence', 
              'silhouette_confidence', 'loudness_confidence', 'whitened_kmeans_confidence',
              'dpcl_classic_confidence'. Defaults to 'silhouette_confidence'.
            kwargs: Keyword arguments to the function being used to compute the confidence.
        """
        if self.model_output is None:
            raise SeparationException(
                "self.model_output is None! Did you run forward?")
        if 'embedding' not in self.model_output:
            raise SeparationException(
                "embedding not in self.model_output! Can't compute confidence.")
        features = self.model_output['embedding']
        if self.metadata['num_channels'] == 1:
            features = features.transpose(0, -2)
        features = features.squeeze(0).transpose(0, 1)
        features = features.cpu().data.numpy()

        confidence_function = getattr(ml.confidence, approach)
        confidence = confidence_function(
            self.audio_signal, features, num_sources, **kwargs)
        return confidence



"""
The effect functions do not augment an AudioSignal object, but rather 
return a FFmpegFilter or a SoXFilter, which may be called on either a sox.transform.Transformer
or a python-ffmpeg stream, depending on the specific effect. To apply the effect on an AudioSignal, 
apply_effect_sox or apply_effect_ffmpeg must be called with the AudioSignal, and a list of
SoXFilters or FFmpegFilters, respectively. 

>>> import nussl.effects
>>> tremolo_filter = effects.tremolo(5, .7)
>>> new_signal = effects.apply_effect_ffmpeg(audio_signal, [tremolo_filter])

Because of this schema, and the requirement that caller must know which effect is a SoX effect
or a FFmpeg effect, we recommend that all users use the AudioSignal hooks for applying effects
rather than calling any functions in this library. 

This line is equivalent to the above snippet

>>> new_signal = audio_signal.tremolo(5, .7).apply_effect()

See also: the associated data augmentation tutorial.
"""

import numpy as np
import warnings
import ffmpeg
import tempfile
import copy
try:
    import soxbindings as sox
except Exception:
    import sox

from .constants import LEVEL_MIN, LEVEL_MAX
from .utils import _close_temp_files


class FilterFunction:
    """
    The FilterFunction class is an abstract class for functions that take 
    audio processing streams, such as ffmpeg-python and pysox effects. 

    Don't call this class. It will not do anything. Please use either FFmpegFilter
    or SoXFilter. 

    Attributes:
        filter: Name of filter used
        params: Parameters for filter
    """
    def __init__(self, filter_, **kwargs):
        self.filter = filter_
        self.params = kwargs

    def __str__(self):
        params = ",".join(f"{p}={v}" for p, v in self.params.items())
        return f"{self.filter} (params: {params})"

    def __call__(self, stream):
        return self.func(stream)

    def func(self, stream):
        pass


class FFmpegFilter(FilterFunction):
    """
    FFmpegFilter is an object returned by FFmpeg effects in effects.py
    To use them, apply_effects_ffmpeg can take a list of effects, and apply them onto an 
    AudioSignal object. 
    """
    def __init__(self, filter_, ffmpeg_name=None, **filter_kwargs):
        super().__init__(filter_, **filter_kwargs)
        ffmpeg_name = self.filter if ffmpeg_name is None else ffmpeg_name
        self.func = lambda stream: stream.filter(ffmpeg_name, **filter_kwargs)


def apply_effects_ffmpeg(audio_signal, filters, silent=False):
    """
    apply_effects_ffmpeg takes an AudioSignal object and a list of FFmpegFilter objects
    and sequentially applies each filter to the signal. 
    Args:
        audio_signal (AudioSignal): AudioSignal object
        filters(list): List of FFmpegFilter objects
        silent (bool): If True, suppresses all FFmpeg output. If False, FFmpeg will log information
        with loglevel 'info'
    Returns:
        augmented_signal(AudioSignal): A new AudioSignal object, with the audio data from 
        audio_signal after applying filters

    """
    # lazy load to avoid circular import
    from . import AudioSignal

    tmpfiles = []

    input_args = {}
    if silent:
        input_args['loglevel'] = 'quiet'

    with _close_temp_files(tmpfiles):
        curr_tempfile = tempfile.NamedTemporaryFile(suffix=".flac")
        out_tempfile = tempfile.NamedTemporaryFile(suffix=".flac")

        tmpfiles.append(curr_tempfile)
        tmpfiles.append(out_tempfile)
        audio_signal.write_audio_to_file(curr_tempfile)

        tmpfiles.append(curr_tempfile)
        stream = ffmpeg.input(curr_tempfile.name, **input_args)
        for filter_ in filters:
            stream = filter_(stream)
        (stream
         .output(out_tempfile.name)
         .overwrite_output()
         .run())

        augmented_data = AudioSignal(path_to_input_file=out_tempfile.name).audio_data

    augmented_signal = audio_signal.make_copy_with_audio_data(augmented_data)
    augmented_signal._effects_applied += filters
    return augmented_signal


class SoXFilter(FilterFunction):
    """
    SoXFilter is an object returned by SoX effects in effects.py
    To use them, apply_effects_sox can take a list of effects, and apply them onto an 
    AudioSignal object. 
    """
    def __init__(self, filter_, **filter_kwargs):
        super().__init__(filter_, **filter_kwargs)
        if filter_ == "time_stretch":
            self.func = lambda tfm: tfm.tempo(**filter_kwargs)
        elif filter_ == "pitch_shift":
            self.func = lambda tfm: tfm.pitch(**filter_kwargs)
        else:
            raise ValueError("Unknown SoX effect passed")


def apply_effects_sox(audio_signal, filters):
    """
    apply_effects_sox takes an AudioSignal object and a list of SoXFilter objects
    and sequentially applies each filter to the signal. 
    Args:
        audio_signal (AudioSignal): AudioSignal object
        filters (list): List of SoXFilter objects
    Returns:
        augmented_signal (AudioSignal): A new AudioSignal object, with the audio data from 
        audio_signal after applying filters
    """
    # lazy load to avoid circular import
    from . import AudioSignal

    audio_data = audio_signal.audio_data

    tfm = sox.Transformer()
    for filter_ in filters:
        tfm = filter_(tfm)
    augmented_data = tfm.build_array(
        input_array=audio_data.T,
        sample_rate_in=audio_signal.sample_rate
    ) 

    augmented_signal = audio_signal.make_copy_with_audio_data(augmented_data)
    augmented_signal._effects_applied += filters
    return augmented_signal


def make_arglist_ffmpeg(lst, sep="|"):
    return sep.join([str(s) for s in lst])


def time_stretch(factor, **kwargs):
    """
    Returns a SoXFilter, when called on an pysox stream, will multiply the 
    tempo of the audio by factor. A factor greater than one will shorten the signal, 
    a factor less then one will lengthen the signal, and a factor of 1 will not change the signal.

    It is recommended that users use `AudioSignal.time_stretch` rather than this function.

    This is a SoX effect. Please see 
    https://pysox.readthedocs.io/en/latest/_modules/sox/transform.html#Transformer.tempo
    for details.
    Args: 
        factor (float): Scaling factor for tempo change. Must be positive.
        kwargs: Arugments passed to `sox.transform.tempo`
    Returns:
        filter (SoXFilter): A SoXFilter object, to be called on an pysndfx stream
    """
    if not np.issubdtype(type(factor), np.number) or factor <= 0:
        raise ValueError("stretch_factor must be a positve scalar")

    return SoXFilter("time_stretch", factor=factor, **kwargs)


def pitch_shift(n_semitones, **kwargs):
    """
    Returns a SoXFilter, when called on an pysox stream, will increase the pitch 
    of the audio by a number of semitones, denoted in n_semitones. A positive n_semitones 
    will raise the pitch of the signal.

    It is recommended that users use `AudioSignal.pitch_shift` rather than this function.

    This is a SoX effect. Please see
    https://pysox.readthedocs.io/en/latest/_modules/sox/transform.html#Transformer.pitch
    for details.

    Args: 
        n_semitones (float): The number of semitones to shift the audio.
            Positive values increases the frequency of the signal
        kwargs: Arugments passed to `sox.transform.pitch`
    Returns:
        filter (SoxFilter): A SoXFilter object, to be called on an pysndfx stream
    """
    return SoXFilter("pitch_shift", n_semitones=n_semitones, **kwargs)


def _pass_arg_check(freq, poles, width_type, width):
    if not np.issubdtype(type(freq), np.number) or freq <= 0:
        raise ValueError("lowest_freq should be positive scalar")
    if poles not in {1, 2}:
        raise ValueError("poles must be either 1 or 2")
    if width_type not in {"h", "q", "o", "s", "k"}:
        raise ValueError("width_type must be either h, q, o, s, or k.")
    if not np.issubdtype(type(width), np.number) or width <= 0:
        raise ValueError("width should be positive scalar")


def low_pass(freq, poles=2, width_type="h", width=.707, **kwargs):
    """
    Creates an FFmpegFilter object, which when called on an ffmpeg stream,
    applies a low pass filter to the audio signal.

    It is recommended that users use `AudioSignal.low_pass` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#lowpass
    for details.
    Args: 
        freq (float): Threshold for low pass. Should be positive
        poles (int): Number of poles. should be either 1 or 2
        width_type (str): Unit of width for filter. Must be either:
            'h': Hz
            'q': Q-factor
            'o': octave
            's': slope
            'k': kHz
        width (float): Band width in width_type units
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    _pass_arg_check(freq, poles, width_type, width)

    filter_ = FFmpegFilter("low_pass", ffmpeg_name="lowpass", f=freq, p=poles,
                           t=width_type, w=width, **kwargs)
    filter_.params = {"freq": freq, "poles": poles,
                      "width_type": width_type, "width": width, **kwargs}
    
    return filter_


def high_pass(freq, poles=2, width_type="h", width=.707, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a high pass filter to the audio signal.

    It is recommended that users use `AudioSignal.high_pass` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#highpass
    for details.
    Args: 
        freq (float): Threshold for high pass. Should be positive scalar
        poles (int): Number of poles. should be either 1 or 2
        width_type (str): Unit of width for filter. Must be either:
            'h': Hz
            'q': Q-factor
            'o': octave
            's': slope
            'k': kHz
        width (float): Band width in width_type units
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    _pass_arg_check(freq, poles, width_type, width)

    filter_ = FFmpegFilter("high_pass", ffmpeg_name="highpass", f=freq, p=poles,
                           t=width_type, w=width, **kwargs)
    filter_.params = {"freq": freq, "poles": poles,
                      "width_type": width_type, "width": width, **kwargs}
    return filter_


def tremolo(mod_freq, mod_depth, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a tremolo filter to the audio signal

    It is recommended that users use `AudioSignal.tremolo` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#tremolo
    for details.
    Args: 
        mod_freq (float): Modulation frequency. Must be between .1 and 20000.
        mod_depth (float): Modulation depth. Must be between 0 and 1.
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    if not np.issubdtype(type(mod_freq), np.number) or mod_freq < .1 or mod_freq > 20000:
        raise ValueError("mod_freq should be positive scalar between .1 and 20000")

    if not np.issubdtype(type(mod_depth), np.number) or mod_depth < 0 or mod_depth > 1:
        raise ValueError("mod_depth should be positve scalar between 0 and 1.")

    filter_ = FFmpegFilter("tremolo", f=mod_freq, d=mod_depth, **kwargs)
    filter_.params = {"mod_freq": mod_freq, "mod_depth": mod_depth, **kwargs}
    return filter_


def vibrato(mod_freq, mod_depth, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a vibrato filter to the audio signal.

    It is recommended that users use `AudioSignal.vibrato` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#vibrato
    for details.
    Args: 
        mod_freq (float): Modulation frequency. Must be between .1 and 20000.
        mod_depth (float): Modulation depth. Must be between 0 and 1.
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    if not np.issubdtype(type(mod_freq), np.number) or mod_freq < .1 or mod_freq > 20000:
        raise ValueError("mod_freq should be positve scalar between .1 and 20000")

    if not np.issubdtype(type(mod_depth), np.number) or mod_depth < 0 or mod_depth > 1:
        raise ValueError("mod_depth should be positve scalar between 0 and 1.")

    filter_ = FFmpegFilter("vibrato", f=mod_freq, d=mod_depth, **kwargs)
    filter_.params = {"mod_freq": mod_freq, "mod_depth": mod_depth, **kwargs}
    return filter_


def chorus(delays, decays, speeds, depths,
           in_gain=.4, out_gain=.4, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a vibrato filter to the audio signal.

    It is recommended that users use `AudioSignal.chorus` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#chorus
    for details.

    Args:
        delays (list of float): delays in ms. Typical Delay is 40ms-6ms
        decays (list of float): decays. Must be between 0 and 1
        speeds (list of float): speeds. Must be between 0 and 1
        depths (list of float): depths. Must be between 0 and 1
        in_gain (float): Proportion of input gain. Must be between 0 and 1
        out_gain (float): Proportion of output gain. Must be between 0 and 1
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter: Resulting filter, to be called on a ffmpeg stream
    """
    if in_gain > 1 or in_gain < 0 or out_gain > 1 or out_gain < 0:
        raise ValueError("in_gain and out_gain must be between 0 and 1.")

    # Bounds could not be found in ffmpeg docs, but recommendations were given
    if min(delays) < 0 or max(delays) > 1000:
        warnings.warn("One or more delays is far from the "
                      "typical 40-60 ms range. This might produce strange results.", UserWarning)

    if (len(delays) != len(decays) or len(decays)
            != len(speeds) or len(speeds) != len(depths)):
        raise ValueError("Delays, decays, depths, and speeds must all be the same length.")

    ffmpeg_delays = make_arglist_ffmpeg(delays)
    ffmpeg_speeds = make_arglist_ffmpeg(speeds)
    ffmpeg_decays = make_arglist_ffmpeg(decays)
    ffmpeg_depths = make_arglist_ffmpeg(depths)

    filter_ = FFmpegFilter("chorus", in_gain=in_gain,
                           out_gain=out_gain, delays=ffmpeg_delays,
                           speeds=ffmpeg_speeds, decays=ffmpeg_decays,
                           depths=ffmpeg_depths, **kwargs)

    filter_.params["delays"] = delays
    filter_.params["speeds"] = speeds
    filter_.params["decays"] = decays
    filter_.params["depths"] = depths
    filter_.params.update(kwargs)
    return filter_


def phaser(in_gain=.4, out_gain=.74, delay=3,
           decay=.4, speed=.5, type_="triangular", **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a phaser filter to the audio signal

    It is recommended that users use `AudioSignal.phaser` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#aphaser
    for details. 
    Args:
        in_gain (float): Proportion of input gain. Must be between 0 and 1
        out_gain (float): Proportion of output gain. Must be between 0 and 1.
        delay (float): Delay of chorus filter in ms. (Time between original signal and delayed)
        decay (float): Decay of copied signal. Must be between 0 and 1.
        speed (float): Modulation speed of the delayed filter. 
        type_ (str): modulation type. Either Triangular or Sinusoidal
            "triangular" or "t" for Triangular
            "sinusoidal" of "s" for sinusoidal
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    if in_gain > 1 or in_gain < 0:
        raise ValueError("in_gain must be between 0 and 1.")
    if out_gain > 1 or out_gain < 0:
        raise ValueError("out_gain must be between 0 and 1.")
    if decay > 1 or decay < 0:
        raise ValueError("delay must be between 0 and 1.")

    allowed_mod_types = {"triangular", "sinusoidal", "t", "s"}
    if type_ not in allowed_mod_types:
        raise ValueError(f"_type must be one of the following:\n{allowed_mod_types}")

    # type is reserved word in python, kwarg dict is necessary
    type_kwarg = {"type": type_}

    filter_ = FFmpegFilter("phaser", ffmpeg_name="aphaser", in_gain=in_gain, out_gain=out_gain,
                           delay=delay, speed=speed, decay=decay, **{**type_kwarg, **kwargs})
    
    filter_.params = copy.deepcopy(filter_.params)
    del filter_.params["type"]
    filter_.params["type_"] = type_
    filter_.params.update(kwargs)
    return filter_
    

def _flanger_argcheck(delay, depth, regen, width,
                      speed, phase, shape, interp):
    error_text = ""
    allowed_shape_types = {"triangular", "sinusoidal"}
    if shape not in allowed_shape_types:
        error_text += f"shape must be one of the following:\n{allowed_shape_types}.\n"
        error_text += f"shape provided is {shape}.\n"
    allowed_interp_types = {"linear", "quadratic"}
    if interp not in allowed_interp_types:
        error_text += f"interp must be one of the following:\n{allowed_interp_types}\n"
        error_text += f"interp provided is {interp}.\n"
    if not 0 <= delay <= 30:
        error_text += f"delay must be in the range 0 <= delay <= 30\n"
        error_text += f"delay provided is {delay}.\n"
    if not -95 <= regen <= 95:
        error_text += f"regen must be in the range -95 <= regen <= 95\n"
        error_text += f"regen provided is {regen}.\n"
    if not 0 <= depth <= 10:
        error_text += f"depth must be in the range 0 <= depth <= 10\n"
        error_text += f"depth provided is {depth}.\n"
    if not .1 <= speed <= 10:
        error_text += f"speed must be in the range .1 <= speed <= 10\n"
        error_text += f"speed provided is {speed}.\n"
    if not 0 <= width <= 100:
        error_text += f"width must be in the range 0 <= width <= 100\n"
        error_text += f"width provided is {width}.\n"
    if not 0 <= phase <= 100:
        error_text += f"phase must be in the range 0 <= phase <= 100\n"
        error_text += f"phase provided is {phase}.\n"
    if error_text:
        raise ValueError(error_text)


def flanger(delay=0, depth=2, regen=0, width=71, speed=.5,
            phase=25, shape="sinusoidal", interp="linear", **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a flanger filter to the audio signal.

    It is recommended that users use `AudioSignal.flanger` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#flanger
    for details. 
    Args:
        delay (float): Base delay in ms between original signal and copy.
            Must be between 0 and 30.
        depth (float): Sweep delay in ms. Must be between 0 and 10.
        regen (float): Percentage regeneration, or delayed signal feedback.
            Must be between -95 and 95.
        width (float): Percentage of delayed signal. Must be between 0 and 100.
        speed (float): Sweeps per second. Must be in .1 to 10
        shape (str): Swept wave shape, Must be "triangular" or "sinusoidal".
        phase (float): swept wave percentage-shift for multi channel. Must be between 0 and 100.
        interp (str): Delay Line interpolation. Must be "linear" or "quadratic".
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """

    _flanger_argcheck(delay, depth, regen, width, speed, phase, shape, interp)

    return FFmpegFilter("flanger", delay=delay, depth=depth, regen=regen, width=width,
                        speed=speed, phase=phase, shape=shape, interp=interp, **kwargs)


def emphasis(level_in, level_out, type_="col", mode='production', **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a emphasis filter to the audio signal. An emphasis filter boosts frequency ranges 
    the most suspectible to noise in a medium. When restoring sounds from such a medium, a 
    de-emphasis filter is used to de-boost boosted frequencies. 

    It is recommended that users use `AudioSignal.emphasis` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#aemphasis
    Args:
        level_in (float): Input gain
        level_out (float): Output gain
        type_ (str): physical medium type to convert/deconvert from.
            Must be one of the following: 
            - "col": Columbia 
            - "emi": EMI
            - "bsi": BSI (78RPM)
            - "riaa": RIAA
            - "cd": CD (Compact Disk)
            - "50fm": 50µs FM
            - "75fm": 75µs FM 
            - "50kf": 50µs FM-KF 
            - "75kf": 75µs FM-KF 
        mode (str): Filter mode. Must be one of the following:
            - "reproduction": Apply de-emphasis filter
            - "production": Apply emphasis filter
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpeg Filter): Resulting filter, to be called on a ffmpeg stream
    """
    if level_in < LEVEL_MIN or level_in > LEVEL_MAX \
            or level_out < LEVEL_MIN or level_out > LEVEL_MAX:
        raise ValueError(f"level_in and level_out must both be between {LEVEL_MIN} AND {LEVEL_MAX}")

    allowed_types = {"col", "emi", "bsi", "riaa", "cd",
                     "50fm", "75fm", "50kf", "75kf"}
    if type_ not in allowed_types:
        raise ValueError(f"Given emphasis filter type is not supported by ffmpeg")
    if mode != "production" and mode != "reproduction":
        raise ValueError(f"mode must be production or reproduction")

    # type is a reserved word in python, so kwarg dict is necessary
    type_kwarg = {'type': type_}

    filter_ = FFmpegFilter("emphasis", ffmpeg_name="aemphasis", level_in=level_in,
                           level_out=level_out, mode=mode, **{**type_kwarg, **kwargs})

    filter_.params = copy.deepcopy(filter_.params)
    del filter_.params["type"]
    filter_.params["type_"] = type_
    filter_.params.update(kwargs)
    return filter_


def _compressor_argcheck(level_in, mode, reduction_ratio,
                         attack, release, makeup, knee, link,
                         detection, mix, threshold):
    error_text = ""
    allowed_mode_types = {"upward", "downward"}
    if mode not in allowed_mode_types:
        error_text += f"shape must be one of the following:\n{allowed_mode_types}.\n"
        error_text += f"shape provided is {mode}.\n"
    allowed_link_types = {"average", "maximum"}
    if link not in allowed_link_types:
        error_text += f"link must be one of the following:\n{allowed_link_types}.\n"
        error_text += f"link provided is {link}.\n"
    allowed_detection_types = {"peak", "rms"}
    if detection not in allowed_detection_types:
        error_text += f"detection must be one of the following:\n{allowed_detection_types}.\n"
        error_text += f"detection provided is {detection}.\n"
    if not LEVEL_MIN <= level_in <= LEVEL_MAX:
        error_text += f"level_in must be in the range {LEVEL_MIN} <= phase <= {LEVEL_MAX}\n"
        error_text += f"level_in provided is {level_in}.\n"
    if not 1 <= reduction_ratio <= 20:
        error_text += f"reduction_ratio must be in the range 1 <= reduction_ratio <= 20\n"
        error_text += f"reduction_ratio provided is {reduction_ratio}.\n"
    if not .01 <= attack <= 2000:
        error_text += f"attack must be in the range 1 <= reduction_ratio <= 2000\n"
        error_text += f"attack provided is {reduction_ratio}.\n"
    if not .01 <= release <= 9000:
        error_text += f"release must be in the range 1 <= release <= 9000\n"
        error_text += f"release provided is {release}.\n"
    if not .1 <= makeup <= 64:
        error_text += f"makeup must be in the range 1 <= makeup <= 64\n"
        error_text += f"makeup provided is {makeup}.\n"
    if not 1 <= knee <= 8:
        error_text += f"knee must be in the range 1 <= knee <= 8\n"
        error_text += f"knee provided is {knee}.\n"
    if not 0 <= mix <= 1:
        error_text += f"mix must be in the range 0 <= mix <= 1\n"
        error_text += f"mix provided is {mix}.\n"
    if not 9.7563e-5 <= threshold <= 1:
        error_text += f"threshold must be in the range 0.000097563 <= threshold <= 1\n"
        error_text += f"threshold provided is {threshold}.\n"
    if error_text:
        raise ValueError(error_text)
    

def compressor(level_in, mode="downward", reduction_ratio=2,
               attack=20, release=250, makeup=1, knee=2.8284, link="average",
               detection="rms", mix=1, threshold=.125, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a compressor filter to the audio signal.

    It is recommended that users use `AudioSignal.compressor` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#acompressor
    Args:
        level_in (float): Input Gain
        mode (str): Mode of compressor operation. Can either be "upward" or "downward". 
        threshold (float): Volume threshold. If a signal's volume is above the threshold,
            gain reduction would apply.
        reduction_ratio (float): Ratio in which the signal is reduced.
        attack (float): Time in ms between when the signal rises above threshold and when 
            reduction is applied
        release (float): Time in ms between when the signal fall below threshold and 
            when reduction is decreased.
        makeup (float): Factor of amplification post-processing
        knee (float): Softens the transition between reduction and lack of thereof. 
            Higher values translate to a softer transition. 
        link (str): Choose average between all channels or mean. String of either
            "average" or "mean.
        detection (str): Whether to process exact signal of the RMS of nearby signals. 
            Either "peak" for exact or "rms".
        mix (float): Proportion of compressed signal in output.
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter (FFmpegFilter): Resulting filter, to be called on a ffmpeg stream
    """
    _compressor_argcheck(level_in, mode, reduction_ratio,
                         attack, release, makeup, knee, link, detection, mix, threshold)

    filter_ = FFmpegFilter("compressor", ffmpeg_name="acompressor", level_in=level_in,
                           ratio=reduction_ratio, attack=attack, release=release, makeup=makeup,
                           knee=knee, link=link, detection=detection, mix=mix, threshold=threshold,
                           **kwargs)
    filter_.params = copy.deepcopy(filter_.params)
    del filter_.params["ratio"]
    filter_.params["reduction_ratio"] = reduction_ratio
    filter_.params.update(**kwargs)
    return filter_


def equalizer(bands, **kwargs):
    """
    Creates a FFmpegFilter object, when called on an ffmpeg stream,
    applies a equalizer filter to the audio signal.

    It is recommended that users use `AudioSignal.equalizer` rather than this function.

    This is a FFmpeg effect. Please see
    https://ffmpeg.org/ffmpeg-all.html#anequalizer
    Args:
        bands (list of dict): A list of dictionaries, for each band. The required values
            for each dictionary:
            'chn' (list of int): List of channel numbers to apply filter. Must be list of ints.
            'f' (float): central freqency of band
            'w' (float): Width of the band in Hz
            'g' (float): Band gain in dB
        A dictionary may also contain these values
            't' (int): Set filter type for band. Default 0. Cann be:
                0, for Butterworth
                1, for Chebyshev type 1
                2, for Chebyshev type 2
        kwargs: Arguments passed to `ffmpeg.filter`
    Returns:
        filter: Resulting filter, to be called on a ffmpeg stream
    """
    for band in bands:
        for chn in band["chn"]:
            if chn < 0 or not np.issubdtype(type(chn), np.integer):
                raise ValueError("All values in band[\"chn\"] must be positive integers")
        if band["f"] <= 0:
            raise ValueError("band[\"f\"] must be a positive scalar")
        if band["w"] <= 0:
            raise ValueError("band[\"w\"] must be a positive scalar")
        if band["g"] <= 0:
            raise ValueError("band[\"g\"] must be a positive scalar")
        if band.get("t", 0) not in {0, 1, 2}:
            raise ValueError("band[\"t\"] must be in {0, 1, 2}")

    params = make_arglist_ffmpeg([
        make_arglist_ffmpeg([
            f"c{c} f={band['f']} w={band['w']} g={band['g']} t={band.get('t', 0)}"
            for c in band["chn"]
        ])
        for band in bands
    ])

    filter_ = FFmpegFilter("equalizer", ffmpeg_name="anequalizer", params=params, **kwargs)
    filter_.params = {"bands": bands, **kwargs}
    return filter_



"""
The :class:`SoftMask` class is for creating a time-frequency mask with values in the range ``[0.0, 1.0]``. Like all 
:class:`separation.masks.mask_base.MaskBase` objects, :class:`SoftMask` is initialized with a 2D or 3D numpy array
containing the mask data. The data type (numpy.dtype) of the initial mask must be float. 
The mask is stored as a 3-dimensional boolean-valued numpy array.

:class:`SoftMask` (like :class:`separation.masks.soft_mask.BinaryMask`) is one of the return types for the :func:`run()` 
methods of :class:`separation.mask_separation_base.MaskSeparationBase`-derived objects (this is most of the 
separation methods in `nussl`.

See Also:
    * :class:`separation.masks.mask_base.MaskBase`: The base class for BinaryMask and SoftMask
    * :class:`separation.masks.soft_mask.BinaryMask`: Similar to BinaryMask, but instead of taking floats, 
      it accepts boolean values.
    * :class:`separation.mask_separation_base.MaskSeparationBase`: Base class for all mask-based separation methods 
      in `nussl`.

Examples:
    Initializing a mask from a numpy array...
    
.. code-block:: python
    :linenos:
    
    import nussl
    import numpy as np
    
    # load a file
    signal = nussl.AudioSignal('path/to/file.wav')
    stft = signal.stft()
    
    # Make a random soft mask with the same shape as the stft
    rand_bool_mask = np.random.random(size=stft.shape)
    bin_mask_bool = nussl.SoftMask(rand_bool_mask)
    
    
:class:`separation.mask_separation_base.MaskSeparationBase`-derived methods return 
:class:`separation.masks.mask_base.MaskBase` masks, like so...

.. code-block:: python
    :linenos:

    import nussl
    
    # load a file
    signal = nussl.AudioSignal('path/to/file.wav')
    
    repet = nussl.Repet(signal, mask_type=nussl.SoftMask)  # You have to specify that you want Binary Masks back
    assert isinstance(repet, nussl.MaskSeparationBase)  # Repet is a MaskSeparationBase-derived class
    
    [background_mask, foreground_mask] = repet.run()  # MaskSeparationBase-derived classes return MaskBase objects
    assert isinstance(foreground_mask, nussl.SoftMask)  # this is True
    assert isinstance(background_mask, nussl.SoftMask)  # this is True
"""

import numpy as np

from . import MaskBase
from . import BinaryMask


class SoftMask(MaskBase):
    """
    A simple class for making a soft mask. The soft mask is represented as a numpy array of floats between
    0.0 and 1.0, inclusive. 
    
    Args:
        input_mask (:obj:`np.ndarray`): 2- or 3-D :obj:`np.array` that represents the mask.
    """

    def __init__(self, input_mask=None, mask_shape=None):
        super(SoftMask, self).__init__(input_mask, mask_shape)

    @staticmethod
    def _validate_mask(mask_):
        assert isinstance(mask_, np.ndarray), 'Mask must be a numpy array!'

        if mask_.dtype.kind not in np.typecodes['AllFloat']:
            raise ValueError('Mask must have type: float! Maybe you want BinaryMask?')

        return mask_

    def mask_to_binary(self, threshold=0.5):
        """
        Create a new :class:`separation.masks.soft_mask.BinaryMask` object from this object's data.
        
        Args:
            threshold (float, Optional): Threshold (between ``[0.0, 1.0]``) to set the True/False cutoff for the binary
             mask.

        Returns:
            A new :class:`separation.masks.soft_mask.BinaryMask` object

        """
        return BinaryMask(self.mask > threshold)

    def invert_mask(self):
        """
        Returns a new mask with inverted values set like ``1 - mask`` for :attr:`mask`.

        Returns:
            A new :class:`SoftMask` object with values set at ``1 - mask``.

        """
        return SoftMask(np.abs(1 - self.mask))



from itertools import permutations, combinations

import numpy as np

from .. import AudioSignal
from ..core import utils


class EvaluationBase(object):
    """
    Base class for all Evaluation classes for source separation algorithms in nussl. 
    Contains common functions for all evaluation techniques. This class should not be 
    instantiated directly.
    
    Both ``true_sources_list`` and ``estimated_sources_list`` get validated 
    using the private method :func:`_verify_input_list`. If your evaluation 
    needs to verify that input is set correctly (recommended) overwrite that method 
    to add checking.
    
    Args:
        true_sources_list (list): List of objects that contain one ground truth source per object.
            In some instances (such as the :class:`BSSEval` objects) this list is filled with
            :class:`AudioSignals` but in other cases it is populated with
            :class:`MaskBase` -derived objects (i.e., either a :class:`BinaryMask` or
            :class:`SoftMask` object).
        estimated_sources_list (list): List of objects that contain source estimations from a source
            separation algorithm. List should be populated with the same type of objects and in the
            same order as :param:`true_sources_list`.
        source_labels (list): List of strings that are labels for each source to be used as keys for
            the scores. Default value is `None` and in that case labels use the file_name attribute.
            If that is also `None`, then the source labels are `Source 0`, `Source 1`, etc.
        compute_permutation (bool): Whether or not to evaluate in a permutation-invariant 
            fashion, where the estimates are permuted to match the true sources. Only the 
            best permutation according to ``best_permutation_key`` is returned to the 
            scores dict. Defaults to False.
        best_permutation_key (str): Which metric to use to decide which permutation of 
            the sources was best.
        **kwargs (dict): Any additional keyword arguments are passed on to ``evaluate_helper``.
    """

    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None, 
                 compute_permutation=False, best_permutation_key=None, **kwargs):
        self.true_sources_list = self._verify_input_list(true_sources_list)
        self.estimated_sources_list = self._verify_input_list(estimated_sources_list)
        self.keys = None

        _source_labels = []
        for i, x in enumerate(self.true_sources_list):
            if isinstance(x, AudioSignal) and x.path_to_input_file:
                label = x.path_to_input_file
            else:
                label = f'source_{i}'
            _source_labels.append(label)

        if source_labels:
            for i, label in enumerate(source_labels):
                _source_labels[i] = label

        self.source_labels = _source_labels
        self.eval_args = kwargs
        self.evaluation_object_type = type(self.true_sources_list[0])
        self._scores = {}
        self.num_channels = self.true_sources_list[0].num_channels
        self.compute_permutation = compute_permutation
        self.best_permutation_key = best_permutation_key

    @staticmethod
    def _verify_input_list(audio_signal_list):
        """
        Base method for verifying a list of input objects for an :class:`EvaluationBase`-derived
        object. Override this method when creating new :class:`EvaluationBased`-derived class.
        
        By default calls :func:`nussl.utils.verify_audio_signal_list_strict`, which verifies that
        all objects in the input list are :class:`audio_signal.AudioSignal` objects with the same
        length, sample rate and have identical number of channels.
        
        Args:
            audio_signal_list (list): List of objects that contain one signal per object. In some
                instances (such as the :class:`BSSEval`) this list is filled with
                :ref:`AudioSignals` but in other cases it is populated with :class:`MaskBase`
                -derived objects (i.e., either a :class:`BinaryMask` or :class:`SoftMask`
                object). In the latter case, this method is overridden with a specific function 
                in :class:`evaluation.precision_recall_fscore.PrecisionRecallFScore`.

        Returns:
            A verified list of objects that are ready for running the evaluation method.

        """
        return utils.verify_audio_signal_list_strict(audio_signal_list)

    def preprocess(self):
        """
        Takes the objects contained in `true_sources_list` and `estimated_sources_list`
        and processes them into numpy arrays that have shape 
        (..., n_channels, n_sources).

        Returns:
            references, estimates in that order as np.ndarrays.

        Note:
            Make sure to return the preprocessed data in the order 
            (references, estimates)!
        """
        raise NotImplementedError('Must implement preprocess in subclass!')
    
    def get_candidates(self):
        """
        This gets all the possible candidates for evaluation. If `compute_permutation`
        is False, then the estimates and the references are assumed to be in the same
        order. The first N estimates will be compared to the first N references, where
        N is min(len(estimates), len(references)).

        If `compute_permutation` is True, and `len(estimates) == len(references)`, then
        every possible ordering of the estimates will be tried to match to the references.
        So if there are 3 references and 3 estimates, a total of 3! = 6 candidates will
        be generated.

        If `compute_permutation` is True and `len(estimates) > len(references)`, then
        every combination of size `len(references)` estimates will be tried as well
        as their permutations. If there are 2 references and 4 estimates, then 
        (4 choose 2) = 6 combos will be tried. For each of those pairs of 2, there will
        be 2! = 2 permutations. So a total of 12 candidates will be generated.

        Returns:
            Two lists of combinations and permutations that should be tried. Each element
            of the list contains the indices that are used to find the sources that
            are compared to each other.

        """
        num_sources = len(self.true_sources_list)
        num_estimates = len(self.estimated_sources_list)

        if not self.compute_permutation:
            return [list(range(num_estimates))], [list(range(num_estimates))]

        combos = list(combinations(range(num_estimates), num_sources))
        orderings = list(permutations(range(num_sources)))
        return combos, orderings

    def evaluate_helper(self, references, estimates, **kwargs):
        """
        This function should be implemented by each class that inherits this
        class. The function should take in a numpy array containing the references and 
        one for the estimates and compute evaluation measures between the two arrays. 
        The results should be stored in a list of dictionaries. For example, a 
        BSSEval evaluator may return a dictionary as follows, for a single estimate:

        .. code-block:: none

                           #or windows or both
            [   {          #ch0  ch1   # results for first estimate
                    "SDR": [5.6, 5.2], # metric
                    "SIR": [9.2, 8.9], # metric
                    "SAR": [4.1, 4.3]  # metric
                }, 
                ...                    # more results for other estimates
            ]

        Each metric should be a key in the dictionary pointing to a value which is a
        list. The list will contain the metrics for however the algorithm was implemented 
        (e.g. there might be two value, one for each channel in a stereo mix, or there
        might be a sequence, one for each window that was evaluated.)
        
        Args:
            references (np.ndarray): References kept in a numpy array. Should have shape
                (..., n_channels, n_sources).
            estimates (np.ndarray): Estimates kept in whatever format you want. Should have
                shape (..., n_channels, n_sources).
            kwargs (dict): Keyword arguments with any additional arguments to be used in 
                the function (e.g. window_size, hop_length).
        
        Returns:
            A list of dictionary containing the measures corresponding to each estimate
            and reference.
        """

        raise NotImplementedError('Must implement evaluate_helper in a subclass!')

    def evaluate(self):
        """
        This function encapsulates the main functionality of all evaluation classes.
        It performs the following steps, some of which must be implemented in subclasses
        of EvaluationBase.

            1. Preprocesses the data somehow into numpy arrays that get passed into your
               evaluation function.
            2. Gets all possible candidates that will be evaluated in your evaluation function.
            3. For each candidate, runs the evaluation function (must be implemented in subclass).
            4. Finds the results from the best candidate.
            5. Returns a dictionary containing those results.

        Steps 1 and 3 must be implemented by the subclass while the others are implemented
        by EvaluationBase.
        
        Returns:
            A dictionary containing the scores for each source for the best candidate.

        """
        references, estimates = self.preprocess()
        combos, orderings = self.get_candidates()
        
        best_permutation_key = self.best_permutation_key
        scores = []
        metrics = []

        for k, combo in enumerate(combos):
            _estimates = estimates[..., combo]
            for o, order in enumerate(orderings):
                _scores = self.evaluate_helper(
                    references[..., order], _estimates, **self.eval_args)

                if not best_permutation_key or best_permutation_key not in _scores[0]:
                    best_permutation_key = sorted(_scores[0].keys())[0]

                _metrics = []
                for _score in _scores:
                    _metrics.extend(_score[best_permutation_key])

                metrics.append(np.mean(_metrics))
                scores.append((combo, order, _scores))
            
        best_idx = np.argmax(metrics)
        combo, order, score = scores[best_idx]
        results = {
            'combination': combo,
            'permutation': order
        }

        for i, o in enumerate(order):
            results[self.source_labels[o]] = score[i]
        self._scores = results

        return results

    @property
    def scores(self):
        """
        A dictionary that stores all scores from the evaluation method. Gets populated when
        :func:`evaluate` gets run.

        """
        return self._scores


class AudioSignalListMismatchError(Exception):
    """
    Error class for when true_sources_list and estimated_sources_list are different
    lengths.
    """
    pass



import torch
import torch.nn as nn
import numpy as np
import gpytorch

class GaussianMixtureTorch(nn.Module):
    def __init__(self, n_components, n_iter=5, covariance_type='diag',
                 covariance_init=1.0, reg_covar=1e-4):
        """
        Initializes a Gaussian mixture model with n_clusters. 
        
        Args:
            n_components (int): Number of components.
            n_iter (int, optional): Number of EM iterations. Defaults to 5.
            covariance_type (str, optional): Covariance type. 
            String describing the type of covariance parameters to use.

            Must be one of:
            'full'
                each component has its own general covariance matrix (this case
                is harder to fit in EM than the others and isn't recommended at
                the moment
            'tied'
                all components share the same general covariance matrix
            'diag'
                each component has its own diagonal covariance matrix
            'spherical'
                each component has its own single variance

            Defaults to 'diag'.

            covariance_init (float, optional): Initial covariance for all
            features and all clusters. Defaults to 0.1.

            reg_covar (float, optional): Regularization amount to add to
            covariance matrix.
        """
        self.n_components = n_components
        self.n_iter = n_iter
        self.covariance_type = covariance_type
        self.covariance_init = covariance_init
        self.reg_covar = reg_covar

        super().__init__()

    def _m_step(self, X, resp):
        """
        Takes a maximization step on the data X.
        
        Args:
            X (torch.Tensor): Data, shape (n_batch, n_samples, n_features)
            resp (torch.Tensor): Responsibilities each Gaussian has for 
            each sample. (n_batch, n_samples, n_components)
        """
        n_batch, n_samples, n_features = X.shape
        _, _, n_components = resp.shape
        resp = resp.view(n_batch, n_samples, n_components, 1)
        X = X.view(n_batch, n_samples, 1, n_features)

        # update means
        _top = (resp * X).sum(dim=1, keepdims=True)
        _bottom = resp.sum(dim=1, keepdims=True)
        means = _top / _bottom

        # update covariance
        diff = X - means
        diff = diff * resp
        covariance = diff.permute(0, 2, 3, 1) @ diff.permute(0, 2, 1, 3)
        covariance = covariance.unsqueeze(1) / _bottom[..., None]
        covariance = covariance.squeeze(1).clamp(min=self.reg_covar)
        covariance = self._enforce_covariance_type(covariance)

        # update prior
        prior = _bottom

        return means.squeeze(1), covariance, prior

    @staticmethod
    def _e_step(X, means, covariance):
        """
        Takes the expectation of X. Returns the log probability of X under each
        Gaussian in the mixture model.
        
        Args:
            X (torch.Tensor): Data, shape (n_batch, n_samples, n_features)
            means (torch.Tensor): Means, shape (n_batch, n_components, n_features)
            covariance (torch.Tensor): (n_batch, n_components, n_features, n_features)
        """
        n_batch, n_samples, n_features = X.shape
        _, n_components, _ = means.shape

        X = X.view(n_batch, n_samples, 1, n_features)
        means = means.view(n_batch, 1, n_components, n_features)
        covariance = covariance.view(
            n_batch, 1, n_components, n_features, n_features)

        mvn = gpytorch.distributions.MultivariateNormal(
            means, covariance_matrix=covariance
        )
        log_prob = mvn.log_prob(X)
        prob = torch.exp(log_prob) + 1e-8
        resp = nn.functional.normalize(prob, p=1, dim=-1)
        return resp, log_prob

    def _enforce_covariance_type(self, covariance):
        n_features = covariance.shape[-1]
        diag_mask = torch.eye(n_features, device=covariance.device)
        diag_mask = diag_mask.reshape(1, 1, n_features, n_features)
        covariance = covariance * diag_mask

        if 'spherical' in self.covariance_type:
            covariance[..., :, :] = (
                covariance.mean(dim=[-2, -1], keepdims=True)
            )
            covariance = covariance * diag_mask

        if 'tied' in self.covariance_type:
            covariance[..., :, :] = covariance.mean(
                dim=1, keepdims=True
            )
        return covariance

    def init_params(self, X, means=None, covariance=None):
        """
        Initializes Gaussian parameters.
        
        Args:
            X (torch.Tensor): Data, shape (n_batch, n_samples, n_features)
            means (torch.Tensor): Means, shape (n_batch, n_components, n_features). Defaults
            to None.
            covariance (torch.Tensor): (n_batch, n_components, n_features, n_features) 
                or (n_batch, n_components, n_features).
            Defaults to None.
        """
        if means is None:
            sampled = X.new(
                X.shape[0], self.n_components).random_(0, X.shape[1])
            sampled += X.new(np.arange(0, X.shape[0])).unsqueeze(
                1).expand(-1, sampled.shape[1]) * X.shape[1]
            sampled = sampled.long()
            means = torch.index_select(
                X.view(-1, X.shape[-1]), 0, sampled.view(-1)).view(
                X.shape[0], sampled.shape[-1], -1)

        if covariance is None:
            covariance = X.new(
                X.shape[0], self.n_components, X.shape[-1]).fill_(
                self.covariance_init).clone()

        if len(covariance.shape) < 4:
            covariance = covariance.unsqueeze(-1).expand(-1, -1, -1, X.shape[-1])

        covariance = self._enforce_covariance_type(covariance.clone())

        return means, covariance

    def forward(self, data, means=None, covariance=None):
        """
        Does a forward pass of the GMM.
        
        Args:
            data (torch.Tensor): Data, shape is (n_batch, ..., n_features)
            means (torch.Tensor): Means, shape (n_batch, n_components, n_features). 
              Defaults to None.
            covariance (torch.Tensor): (n_batch, n_components, n_features, n_features) 
              or (n_batch, n_components, n_features). 
        
        Returns:
            dict: Containing keys 'resp', 'log_prob', 'means', 'covariance', 'prior'.
        """
        shape = data.shape
        data = data.view(shape[0], -1, shape[-1])
        means, covariance = self.init_params(data, means, covariance)

        resp = log_prob = prior = None
        for i in range(self.n_iter):
            resp, log_prob = self._e_step(data, means, covariance)
            means, covariance, prior = self._m_step(data, resp)

        return {
            'resp': resp.view(shape[:-1] + (-1,)),
            'log_prob': log_prob.view(shape[:-1] + (-1,)),
            'means': means,
            'covariance': covariance,
            'prior': prior
        }
